/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>

#include "libdeslite/des.h"

#include "constants.h"
#include "state.h"
#include "packet.h"
#include "defs.h"
#include "log.h"
#include "cookie.h"
#include "timer.h"
#include "rnd.h"
#include "ipsec_doi.h"

#ifndef JI
#include "sha1.h"
#include "md5.h"
#endif

#ifdef JI
#include <linux/skbuff.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_ipe4.h"
#include "ipsec_ah.h"
#include "ipsec_esp.h"
#endif

static u_char *get_preshared_key(struct state *st, int *length);

static void initiate_quick(int sock, struct state *st, struct sockaddr_in sin2,
	       struct sockaddr_in sin3, struct sockaddr_in sin4,
  	       struct sockaddr_in sin5);

/*
 * This file has all the IPsec DOI and Oakley resolution routines.
 */

/*
 * XXX This should be replaced by a call to the kernel when
 * XXX we get an API.
 * The returned SPI is (should be) in network byte order.
 */
static u_int32_t
oakley_get_spi(void)
{
    static u_int32_t spi = 0x100;

    return htonl(spi++);
}

#ifdef JI				/* Support for ipsec-0.5 */
/*
 * Setup an IPsec route entry. Code taken from addrt.c.
 * XXX Doesn't handle proxy (firewall mode) yet; it does
 * XXX encapsulation on our packets only (useful to protect
 * XXX the IP options).
 */
void
setup_route(struct state *st)
{
    int fd;
    struct encap_msghdr emh;
    struct sockaddr_in sin;

    fd = open("/dev/ipsec", O_WRONLY);
    if (fd < 0)
      exit_log("open() failed in setup_route()");

    bcopy(&(st->st_peer), &sin, sizeof(sin));

    emh.em_magic = EM_MAGIC;
    emh.em_msglen = sizeof(emh);
    emh.em_version = 0;
    emh.em_type = EMT_SETEROUTE;
    emh.em_eaddr.sen_len = emh.em_emask.sen_len
			 = sizeof(struct sockaddr_encap);
    emh.em_eaddr.sen_family = emh.em_emask.sen_family = 26;
    emh.em_eaddr.sen_type = SENT_IP4;
    emh.em_emask.sen_type = 255;

#ifdef DEBUG
    log(0, "we're here...");
#endif

    if (st->st_ouruser.sin_addr.s_addr != 0)	/* Proxy -- gateway/firewall */
    {
      	emh.em_eaddr.sen_ip_src.s_addr = st->st_ouruser.sin_addr.s_addr;
	emh.em_eaddr.sen_ip_dst.s_addr = st->st_peeruser.sin_addr.s_addr;
	emh.em_emask.sen_ip_src.s_addr = st->st_ournetmask.sin_addr.s_addr;
	emh.em_emask.sen_ip_dst.s_addr = st->st_peernetmask.sin_addr.s_addr;
    }
    else	/* XXX use our identity -- assumed to be an IP address */
    {
      	emh.em_eaddr.sen_ip_src.s_addr = *((u_long *)st->st_myidentity);
    	emh.em_eaddr.sen_ip_dst.s_addr = sin.sin_addr.s_addr;
    	emh.em_emask.sen_ip_src.s_addr = inet_addr("255.255.255.255");
    	emh.em_emask.sen_ip_dst.s_addr = inet_addr("255.255.255.255");
    }

    emh.em_erdst.s_addr = sin.sin_addr.s_addr;

#ifdef DEBUG
    log(0, "...and here");
#endif

    bcopy(st->st_peer_spi, &(emh.em_erspi), sizeof(emh.em_erspi));
    if (st->st_peer_encapsulation_mode == ENCAPSULATION_MODE_TUNNEL)
      emh.em_erspi = ~emh.em_erspi;

    if (write(fd, &emh, sizeof(emh)) != sizeof(emh))
      exit_log("write() failed in add_route()");

#ifdef DEBUG
    log(0, "route to %s setup", get_address_in(sin));
#endif

    close(fd);
}

/*
 * Setup a pair of SAs. Code taken from setsa.c and spigrp.c, in
 * ipsec-0.5.
 * XXX Just does the 3DES-MD5-96 transform, as a guideline. See
 * XXX ipsec-0.5/utils for code that handles the other transforms.
 */
void
setup_sa(struct state *st)
{
    u_long bigbuf[1024];
    struct encap_msghdr *em = (struct encap_msghdr *)bigbuf;
    struct ipe4_xdata *ipe4xd = (struct ipe4_xdata *)(em->em_dat);
    struct ahmd5_xdata *ahmd5xd = (struct ahmd5_xdata *)(em->em_dat);
    struct ahhmacmd5_edata *hmacmd5xd = (struct ahhmacmd5_edata *)(em->em_dat);
    struct espdes_xdata *desxd = (struct espdes_xdata *)em->em_dat;
    struct espblkrply_edata *blkxd = (struct espblkrply_edata *)(em->em_dat);
    int fd;
    struct sockaddr_in sin;
    u_int32_t tempspi;

    fd = open("/dev/ipsec", O_WRONLY);
    if (fd < 0)
      exit_log("open() failed in setup_sa()");

    bcopy(&(st->st_peer), &sin, sizeof(sin));

    bzero(bigbuf, sizeof(bigbuf));

    em->em_magic = EM_MAGIC;
    /* em->em_msglen is set by the appropriate case */
    em->em_version = 0;
    em->em_type = EMT_SETSPI;
    /* XXX This assumes myidentity is an IP address */
    em->em_dst.s_addr = *((u_long *)st->st_myidentity);
    bcopy(st->st_spi, &(em->em_spi), sizeof(em->em_spi));
    em->em_if = 0;
    /* em->em_alg is set by the appropriate case */

    /* Set our receiving SA */
    switch (st->st_protoid)
    {
	case PROTO_IPSEC_ESP:
	    switch (st->st_transid)
	    {
		case ESP_3DES:
		    blkxd->eme_flags |= EME_INITIATOR;
		    blkxd->eme_flags |= EME_CONSTIV;
		    bcopy(st->st_keymat, blkxd->eme_key, st->st_keymat_len);
		    blkxd->eme_klen = st->st_keymat_len;
		    em->em_msglen = EMT_SETSPI_FLEN +
				    sizeof (struct espblkrply_edata);
		    blkxd->eme_ooowin = 32;
		    em->em_alg = XF_ESP3DESMD596;
		    break;

		default:
		    exit_log("not implemented yet");
	    }
	    break;

	case PROTO_IPSEC_AH:
	    exit_log("not implemented yet");
	    break;
    }

    if (write(fd, (caddr_t)em, em->em_msglen) != em->em_msglen)
      exit_log("write() failed in setup_sa()");
#ifdef DEBUG
    else
      log(0, "sent %d bytes to the kernel", em->em_msglen);
#endif

    bzero(bigbuf, sizeof(bigbuf));

    em->em_magic = EM_MAGIC;
    /* em->em_msglen is set by the appropriate case */
    em->em_version = 0;
    em->em_type = EMT_SETSPI;
    em->em_dst.s_addr = sin.sin_addr.s_addr;

    /*
     * XXX hack alert -- we SHOULD NOT HAVE TO HAVE A DIFFERENT SPI
     * XXX FOR IP-in-IP ENCAPSULATION!
     */
    if (st->st_peer_encapsulation_mode == ENCAPSULATION_MODE_TUNNEL)
    {
  	bcopy(st->st_peer_spi, &tempspi, sizeof(tempspi));
	em->em_spi = tempspi;
	tempspi = ~tempspi;
    }
    else
      bcopy(st->st_peer_spi, &(em->em_spi), sizeof(em->em_spi));

    em->em_if = 0;
    /* em->em_alg is set by the appropriate case */

    /* Set our sending (peer's) SA */
    switch (st->st_peer_protoid)
    {
	case PROTO_IPSEC_ESP:
	    switch (st->st_peer_transid)
	    {
		case ESP_3DES:
		    blkxd->eme_flags |= EME_INITIATOR;
		    blkxd->eme_flags |= EME_CONSTIV;
		    bcopy(st->st_peer_keymat, blkxd->eme_key,
			  st->st_peer_keymat_len);
		    blkxd->eme_klen = st->st_peer_keymat_len;
		    em->em_msglen = EMT_SETSPI_FLEN +
				    sizeof (struct espblkrply_edata);
		    blkxd->eme_ooowin = 32;
		    em->em_alg = XF_ESP3DESMD596;
		    break;

		default:
		    exit_log("not implemented yet");
	    }
	    break;

	case PROTO_IPSEC_AH:
	    exit_log("not implemented yet");
	    break;
    }

    if (write(fd, (caddr_t)em, em->em_msglen) != em->em_msglen)
      exit_log("write() failed in setup_sa()");
#ifdef DEBUG
    else
      log(0, "sent %d bytes to the kernel", em->em_msglen);
#endif

    if (st->st_peer_encapsulation_mode == ENCAPSULATION_MODE_TUNNEL)
    {
#ifdef DEBUG
	log(0, "setting up encapsulation");
#endif

	bzero(bigbuf, sizeof(bigbuf));

	em->em_magic = EM_MAGIC;
	em->em_version = 0;
	em->em_type = EMT_SETSPI;
	em->em_dst.s_addr = sin.sin_addr.s_addr;
	em->em_spi = tempspi;
	em->em_if = 0;
        em->em_msglen = EMT_SETSPI_FLEN + EMT_IPE4_ULEN;
        em->em_alg = XF_IP4;

	/* XXX This assumes that myidentity is an IP address */
        ipe4xd->i4_src.s_addr = *((u_long *)st->st_myidentity);
        ipe4xd->i4_dst.s_addr = sin.sin_addr.s_addr;

	if (write(fd, (caddr_t)em, em->em_msglen) != em->em_msglen)
	  exit_log("write() failed in setup_sa()");
#ifdef DEBUG
	else
	  log(0, "sent %d bytes to the kernel", em->em_msglen);
#endif

	/* Group the two SPIs */
	bzero(bigbuf, sizeof(bigbuf));

	em->em_magic = EM_MAGIC;
	em->em_version = 0;
	em->em_type = EMT_GRPSPIS;
        em->em_msglen = EM_MINLEN + 2 * 12;

	em->em_rel[1].emr_spi = ~tempspi;
	em->em_rel[1].emr_dst.s_addr = sin.sin_addr.s_addr;

	em->em_rel[0].emr_spi = tempspi;
	em->em_rel[0].emr_dst.s_addr = sin.sin_addr.s_addr;

#ifdef DEBUG
	fprintf(stdout, "linking %s/%08x to ", inet_ntoa(sin.sin_addr),
		ntohl(em->em_rel[0].emr_spi));
	fprintf(stdout, "%s/%08x\n", inet_ntoa(sin.sin_addr),
		ntohl(em->em_rel[1].emr_spi));
#endif

	if (write(fd, (caddr_t)em, em->em_msglen) != em->em_msglen)
	  exit_log("write() failed in setup_sa()");
#ifdef DEBUG
	else
	  log(0, "sent %d bytes to the kernel", em->em_msglen);
#endif
    }

    close(fd);
}

#endif /* JI */

/*
 * Send a notification to the peer. We could make a decision on
 * whether to send the notification, based on the type and the
 * destination, if we care to.
 * XXX It doesn't handle DELETE notifications (which are also
 * XXX informational exchanges).
 */
static void
send_notification(int sock, u_int16_t type, u_char *spi,
		  u_char spilen, u_char protoid, u_char *icookie,
		  u_char *rcookie, u_int32_t msgid, struct sockaddr sa)
{
    u_char buffer[sizeof(struct isakmp_hdr) +
		 sizeof(struct isakmp_notification)];
    struct isakmp_hdr *isa = (struct isakmp_hdr *) buffer;
    struct isakmp_notification *isan = (struct isakmp_notification *)
				       (buffer + sizeof(struct isakmp_hdr));

    bzero(buffer, sizeof(struct isakmp_hdr) +
	  sizeof(struct isakmp_notification));

    if (icookie != (u_char *) NULL)
      bcopy(icookie, isa->isa_icookie, COOKIE_SIZE);

    if (rcookie != (u_char *) NULL)
      bcopy(rcookie, isa->isa_rcookie, COOKIE_SIZE);

    /* Standard header */
    isa->isa_np = ISAKMP_NEXT_N;
    isa->isa_maj = ISAKMP_MAJOR_VERSION;
    isa->isa_min = ISAKMP_MINOR_VERSION;
    isa->isa_xchg = ISAKMP_XCHG_INFO;
    isa->isa_msgid = msgid;
    isa->isa_length = htonl(sizeof(struct isakmp_hdr) +
			    sizeof(struct isakmp_notification) +
			    spilen);

    /* Notification header */
    isan->isan_type = htons(type);
    isan->isan_doi = htonl(ISAKMP_DOI_IPSEC);
    isan->isan_length = htons(sizeof(struct isakmp_notification) + spilen);
    isan->isan_spisize = spilen;
    bcopy(spi, (u_char *)isan + sizeof(struct isakmp_notification), spilen);
    isan->isan_protoid = protoid;

#ifdef DEBUG
    log(0, "sending INFO type %d to %s, port %d", type,
	get_address(sa), get_port(sa));
#endif

    if (sendto(sock, buffer, ntohl(isa->isa_length), 0, &sa,
	       sizeof(sa)) != ntohl(isa->isa_length))
      log(1, "sendto() failed in send_notification() to %s, port %d",
	  get_address(sa), get_port(sa));
#ifdef DEBUG
    else
      log(0, "transmitted %d bytes", ntohl(isa->isa_length));
#endif
}

/*
 * Get an msgid for Phase 2 messages.
 */
static u_int32_t
oakley_get_msgid(struct sockaddr sa)
{
    u_int32_t msgid;
    struct messageid *mid;

    msgid = find_messageid(sa);
    if (msgid == 0)
    {
	mid = insert_messageid(sa);
	msgid = mid->md_msgid;
    }

#ifdef DEBUG
          log(0, "MSG-ID is %d", msgid);
#endif

    return msgid;
}

/*
 * Get an ESP Proposal. The second argument could be used to decide on
 * a per-peer basis what the proposal(s) should be.
 * If the fourth argument is true, then we include the encapsulation
 * attribute.
 * XXX For now, we offer one very basic proposal.
 */

static u_char *
oakley_get_esp_prop(int *length, struct sockaddr_in sin, struct state *st,
		    int tunnel)
{
    u_char p[8192], *q;
    int i, proposal = 1, transform = 1, k;
    u_int32_t spi;
    struct isakmp_proposal *isap;
    struct isakmp_transform *isat;
    struct isakmp_attribute *isaa;

    bzero(p, sizeof(p));

    /* Proposal header -- ESP */
    isap = (struct isakmp_proposal *) p;
    i = sizeof(struct isakmp_proposal);

    isap->isap_np = ISAKMP_NEXT_NONE;
    isap->isap_proposal = proposal++;
    isap->isap_protoid = PROTO_IPSEC_ESP;
    isap->isap_spisize = IPSEC_DOI_SPI_SIZE;
    isap->isap_notrans = 1;
    st->st_protoid = isap->isap_protoid;

#ifdef DEBUG
    fprintf(stdout, "Protocol: %d\n", isap->isap_protoid);
#endif

    /* Get an SPI -- it is assumed it's already in network byte order */
    spi = oakley_get_spi();
    bcopy(&spi, p + i, sizeof(spi));

    st->st_spi = (u_char *) calloc(IPSEC_DOI_SPI_SIZE, sizeof(u_char));
    if (st->st_spi == (u_char *) NULL)
      exit_log("calloc() failed in oakley_get_esp_prop()");
    bcopy(&spi, st->st_spi, sizeof(spi));
    st->st_spi_len = IPSEC_DOI_SPI_SIZE;

    i += IPSEC_DOI_SPI_SIZE;

#ifdef DEBUG
    fprintf(stdout, "SPI sent: ");
    for (k = 0; k < st->st_spi_len; k++)
      fprintf(stdout, "%02x ", st->st_spi[k]);
    fprintf(stdout, "\n");
#endif

    /* Transform header -- ESP-DES-MD5 */
    isat = (struct isakmp_transform *) (p + i);

    isat->isat_np = ISAKMP_NEXT_NONE;
    isat->isat_transnum = transform++;
    isat->isat_transid = ESP_3DES;
    isat->isat_length = i;		/* Use as a temporary variable */
    st->st_transid = isat->isat_transid;
    i += sizeof(struct isakmp_transform);

#ifdef DEBUG
    fprintf(stdout, "Transform: %d\n", isat->isat_transid);
#endif

    /* SA lifetime type */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = SA_LIFE_TYPE;

    isaa->isaat_lv = htons(SA_LIFE_TYPE_SECONDS);
    isaa->isaat_af = 1;
    isaa->isaat_type = htons(isaa->isaat_type);
    i += sizeof(struct isakmp_attribute);

#ifdef DEBUG
    fprintf(stdout, "SA lifetime type: %d\n", ntohs(isaa->isaat_lv));
#endif

    /* SA lifetime */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = SA_LIFE_DURATION;
    isaa->isaat_lv = htons(SA_LIFE_DURATION_DEFAULT);
    isaa->isaat_af = 1;
    isaa->isaat_type = htons(isaa->isaat_type);
    st->st_expire = ntohs(isaa->isaat_lv);
    i += sizeof(struct isakmp_attribute);

#ifdef DEBUG
    fprintf(stdout, "SA lifetime: %u\n", ntohs(isaa->isaat_lv));
#endif

    /* Encapsulation mode */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = ENCAPSULATION_MODE;
    if (tunnel)
      isaa->isaat_lv = htons(ENCAPSULATION_MODE_TUNNEL);
    else
      isaa->isaat_lv = htons(ENCAPSULATION_MODE_TRANSPORT);
    isaa->isaat_af = 1;
    isaa->isaat_type = htons(isaa->isaat_type);
    st->st_encapsulation_mode = ntohs(isaa->isaat_lv);
    i += sizeof(struct isakmp_attribute);

#ifdef DEBUG
    fprintf(stdout, "Encapsulation mode: %d\n", ntohs(isaa->isaat_lv));
#endif

    /* AUTH algorithm */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = AUTH_ALGORITHM;
    isaa->isaat_lv = htons(AUTH_ALGORITHM_HMAC_MD5);
    isaa->isaat_af = 1;
    isaa->isaat_type = htons(isaa->isaat_type);
    st->st_hmac_algorithm = ntohs(isaa->isaat_lv);
    i += sizeof(struct isakmp_attribute);

#ifdef DEBUG
    fprintf(stdout, "AUTH algorithm: %d\n", ntohs(isaa->isaat_lv));
#endif

    /* A couple of default values */
    st->st_authkey_duration = SA_LIFE_DURATION_DEFAULT;
    st->st_enckey_duration = SA_LIFE_DURATION_DEFAULT;

    isat->isat_length = htons(i - isat->isat_length);
    isap->isap_length = htons(i);

    q = (u_char *)calloc(i, sizeof(u_char));
    if (q == (u_char *) NULL)
      exit_log("calloc() failed in oakley_get_esp_prop()");

    bcopy(p, q, i);
    *length = i;
    return q;
}

/*
 * Get an AH Proposal.
 * XXX For now, we offer one very basic proposal.
 */
static u_char *
oakley_get_ah_prop(int *length, struct sockaddr_in sin, struct state *st,
		   int tunnel)
{
    u_char p[8192], *q;
    int i, proposal = 1, transform = 1, k;
    u_int32_t spi;
    struct isakmp_proposal *isap;
    struct isakmp_transform *isat;
    struct isakmp_attribute *isaa;

    bzero(p, sizeof(p));

    /* Proposal header -- AH */
    isap = (struct isakmp_proposal *) p;
    i = sizeof(struct isakmp_proposal);

    isap->isap_np = ISAKMP_NEXT_NONE;
    isap->isap_proposal = proposal++;
    isap->isap_protoid = PROTO_IPSEC_AH;
    isap->isap_spisize = IPSEC_DOI_SPI_SIZE;
    isap->isap_notrans = 1;
    st->st_protoid = isap->isap_protoid;

#ifdef DEBUG
    fprintf(stdout, "Protocol: %d\n", isap->isap_protoid);
#endif

    /* Get an SPI -- it is assumed it's already in network byte order */
    spi = oakley_get_spi();
    bcopy(&spi, p + i, sizeof(spi));

    st->st_spi = (u_char *) calloc(IPSEC_DOI_SPI_SIZE, sizeof(u_char));
    if (st->st_spi == (u_char *) NULL)
      exit_log("calloc() failed in oakley_get_ah_prop()");
    bcopy(&spi, st->st_spi, sizeof(spi));
    st->st_spi_len = IPSEC_DOI_SPI_SIZE;

    i += IPSEC_DOI_SPI_SIZE;

#ifdef DEBUG
    fprintf(stdout, "SPI sent: ");
    for (k = 0; k < st->st_spi_len; k++)
      fprintf(stdout, "%02x ", st->st_spi[k]);
    fprintf(stdout, "\n");
#endif

    /* Transform header -- AH-MD5 */
    isat = (struct isakmp_transform *) (p + i);

    isat->isat_np = ISAKMP_NEXT_NONE;
    isat->isat_transnum = transform++;
    isat->isat_transid = AH_MD5;
    isat->isat_length = i;		/* Use as a temporary variable */
    st->st_transid = isat->isat_transid;
    i += sizeof(struct isakmp_transform);

#ifdef DEBUG
    fprintf(stdout, "Transform: %d\n", isat->isat_transid);
#endif

    /* SA lifetime type */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = SA_LIFE_TYPE;
    isaa->isaat_lv = htons(SA_LIFE_TYPE_SECONDS);
    isaa->isaat_af = 1;
    isaa->isaat_type = htons(isaa->isaat_type);
    i += sizeof(struct isakmp_attribute);

#ifdef DEBUG
    fprintf(stdout, "SA lifetime type: %d\n", ntohs(isaa->isaat_lv));
#endif

    /* SA lifetime */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = SA_LIFE_DURATION;
    isaa->isaat_lv = htons(SA_LIFE_DURATION_DEFAULT);
    isaa->isaat_af = 1;
    isaa->isaat_type = htons(isaa->isaat_type);
    st->st_expire = ntohs(isaa->isaat_lv);
    i += sizeof(struct isakmp_attribute);

#ifdef DEBUG
    fprintf(stdout, "SA lifetime: %u\n", ntohs(isaa->isaat_lv));
#endif

    /* Encapsulation mode */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = ENCAPSULATION_MODE;
    if (tunnel)
      isaa->isaat_lv = htons(ENCAPSULATION_MODE_TUNNEL);
    else
      isaa->isaat_lv = htons(ENCAPSULATION_MODE_TRANSPORT);
    isaa->isaat_af = 1;
    isaa->isaat_type = htons(isaa->isaat_type);
    st->st_encapsulation_mode = ntohs(isaa->isaat_lv);
    i += sizeof(struct isakmp_attribute);

    isat->isat_length = htons(i - isat->isat_length);
    isap->isap_length = htons(i);

#ifdef DEBUG
    fprintf(stdout, "Encapsulation mode: %d\n", ntohs(isaa->isaat_lv));
#endif

    /* A few default values */
    st->st_authkey_duration = SA_LIFE_DURATION_DEFAULT;
    st->st_enckey_duration = SA_LIFE_DURATION_DEFAULT;

    q = (u_char *) calloc(i, sizeof(u_char));
    if (q == (u_char *) NULL)
      exit_log("calloc() failed in oakley_get_ah_prop()");

    bcopy(p, q, i);
    *length = i;

    return q;
}

/*
 * Get an Oakley Proposal.
 * XXX For now, we offer one very basic proposal.
 * XXX Since we only do this one proposal, we also copy the value in
 * XXX the state object at this point. This will have to change if/when
 * XXX we start offering multiple proposals (see ipsec_prop_to_state()).
 */
static u_char *
oakley_get_oakley_prop(int *length, struct sockaddr_in sin, struct state *st)
{
    u_char p[8192], *q, *pp;
    int i, proposal = 1, transform = 1, k;
    u_int32_t spi;
    struct isakmp_proposal *isap;
    struct isakmp_transform *isat;
    struct isakmp_attribute *isaa;

    bzero(p, sizeof(p));

    /* Proposal header -- ISAKMP */
    isap = (struct isakmp_proposal *) p;
    i = sizeof(struct isakmp_proposal);

    isap->isap_np = ISAKMP_NEXT_NONE;
    isap->isap_proposal = proposal++;
    isap->isap_protoid = PROTO_ISAKMP;
    isap->isap_spisize = IPSEC_DOI_SPI_SIZE;
    isap->isap_notrans = 1;
    st->st_protoid = PROTO_ISAKMP;

    /* Get an SPI -- it is assumed it's already in network byte order */
    spi = oakley_get_spi();
    bcopy(&spi, p + i, sizeof(spi));

    st->st_spi = (u_char *) calloc(IPSEC_DOI_SPI_SIZE, sizeof(u_char));
    if (st->st_spi == (u_char *) NULL)
      exit_log("calloc() failed in oakley_get_oakley_prop()");
    bcopy(&spi, st->st_spi, sizeof(spi));
    st->st_spi_len = IPSEC_DOI_SPI_SIZE;

    i += IPSEC_DOI_SPI_SIZE;

    /* Transform header -- Oakley */
    isat = (struct isakmp_transform *) (p + i);

    isat->isat_np = ISAKMP_NEXT_NONE;
    isat->isat_transnum = transform++;
    isat->isat_transid = KEY_OAKLEY;
    st->st_transid = KEY_OAKLEY;
    isat->isat_length = i;		/* Use as a temporary variable */
    i += sizeof(struct isakmp_transform);

    /* Encryption algorithm attribute */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = OAKLEY_ENCRYPTION_ALGORITHM;
    isaa->isaat_lv = htons(OAKLEY_DES_CBC);
    isaa->isaat_af = 1;
    st->st_enc = OAKLEY_DES_CBC;
    isaa->isaat_type = htons(isaa->isaat_type);
    i += sizeof(struct isakmp_attribute);

    /* Hash algorithm attribute */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = OAKLEY_HASH_ALGORITHM;
    isaa->isaat_lv = htons(OAKLEY_MD5);
    isaa->isaat_af = 1;
    st->st_hash = OAKLEY_MD5;
    isaa->isaat_type = htons(isaa->isaat_type);
    i += sizeof(struct isakmp_attribute);

    /* Authentication method attribute */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = OAKLEY_AUTHENTICATION_METHOD;

    pp = get_preshared_key(st, &k);
    if (pp != (u_char *) NULL)
    {
        st->st_auth = OAKLEY_PRESHARED_KEY;
        isaa->isaat_lv = htons(OAKLEY_PRESHARED_KEY);
	free(pp);
    }
    else
    {
	/* XXX Find another way to authenticate */
	exit_log("put a shared secret for %s in there, will ya ?",
		 get_address_in(sin));
    }

    isaa->isaat_af = 1;
    isaa->isaat_type = htons(isaa->isaat_type);
    i += sizeof(struct isakmp_attribute);

    /* Group Description attribute */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = OAKLEY_GROUP_DESCRIPTION;
    isaa->isaat_lv = htons(OAKLEY_DEFAULT_GROUP);
    isaa->isaat_af = 1;
    st->st_group = OAKLEY_DEFAULT_GROUP;
    isaa->isaat_type = htons(isaa->isaat_type);
    i += sizeof(struct isakmp_attribute);

    /* Lifetime Type attribute */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = OAKLEY_LIFE_TYPE;
    isaa->isaat_lv = htons(OAKLEY_LIFE_SECONDS);
    isaa->isaat_af = 1;
    isaa->isaat_type = htons(isaa->isaat_type);
    i += sizeof(struct isakmp_attribute);

    /* Lifetime Duration attribute */
    isaa = (struct isakmp_attribute *) (p + i);
    isaa->isaat_type = OAKLEY_LIFE_DURATION;
    isaa->isaat_lv = htons(OAKLEY_ISAKMP_SA_LIFETIME);
    isaa->isaat_af = 1;
    st->st_expire = OAKLEY_ISAKMP_SA_LIFETIME;
    isaa->isaat_type = htons(isaa->isaat_type);
    i += sizeof(struct isakmp_attribute);

    isat->isat_length = htons(i - isat->isat_length);
    isap->isap_length = htons(i);

    q = (u_char *) calloc(i, sizeof(u_char));
    if (q == (u_char *) NULL)
      exit_log("calloc() failed in oakley_get_oakley_prop()");

    bcopy(p, q, i);
    *length = i;

    return q;
}

/*
 * Construct an SA payload for Oakley, depending on the goal. Return
 * a valid SA payload, length of the buffer on the first argument.
 */
static u_char *
oakley_get_sa(int *length, struct sockaddr_in sin, int goal, struct state *st)
{
    struct isakmp_sa *isa;
    u_char p[8192], *q;
    int i, j;

    bzero(p, sizeof(p));

    isa = (struct isakmp_sa *) p;

    isa->isasa_np = ISAKMP_NEXT_NONE;
    isa->isasa_doi = htonl(ISAKMP_DOI_IPSEC);

    i = sizeof(struct isakmp_sa);

    /* We don't do labeled integrity/secrecy, so forget the LDI */
    p[i + IPSEC_DOI_SITUATION_LENGTH - 1] = SIT_IDENTITY_ONLY;
    i += IPSEC_DOI_SITUATION_LENGTH;

    if (goal & GOAL_KEYMANAGEMENT)
    {
	q = oakley_get_oakley_prop(&j, sin, st);
	if (q == (u_char *) NULL)
	  return q;

	bcopy(q, p + i, j);
	free(q);
	i += j;
    }
    else
    {
	if (goal & GOAL_ENCRYPT)	/* Does authentication too */
	{
	    q = oakley_get_esp_prop(&j, sin, st, goal & GOAL_TUNNEL);
	    if (q == (u_char *) NULL)
	      return q;

	    bcopy(q, p + i, j);
	    free(q);
	    i += j;
	}
	else
	  if (goal & GOAL_AUTHENTICATE)
	  {
	      q = oakley_get_ah_prop(&j, sin, st, goal & GOAL_TUNNEL);
	      if (q == (u_char *) NULL)
		return q;

	      bcopy(q, p + i, j);
	      free(q);
	      i += j;
	  }
	  else
	    exit_log("no valid goal specified (%d)", goal); /* XXX */
    }

    isa->isasa_length = htons(i);
    q = (u_char *) calloc(i, sizeof(u_char));
    if (q == (u_char *) NULL)
      exit_log("calloc() failed in oakley_get_sa()");

    bcopy(p, q, i);
    *length = i;

    return q;
}

/*
 * Initiate an Oakley exchange.
 */
static void
oakley_initiate(int sock, struct sockaddr_in sin, struct sockaddr_in sin2,
		struct sockaddr_in sin3, struct sockaddr_in sin4,
		struct sockaddr_in sin5, int goal)
{
    u_char *buffer, *p;
    struct isakmp_hdr isa;
    struct isakmp_sa *isasa;
    struct state *st;
    int j;

    /*
     * If there's already an ISAKMP SA established, use that and
     * go directly to Quick Mode.
     * XXX If an ISAKMP SA is *being* established, we foolishly
     * XXX try to establish another one, in parallel. We could
     * XXX issue an event to wait for it to finish and then try
     * XXX to establish it.
     */
    st = find_phase1_state((struct sockaddr *)&sin);
    if (st != (struct state *) NULL)
    {
#ifdef DEBUG
	log(0, "Doing Quick Mode with %s, port %d, goal %d",
	    get_address_in(sin), get_port_in(sin), goal);
#endif
	st->st_goal = goal;
	initiate_quick(sock, st, sin2, sin3, sin4, sin5);
	return;
    }

    bzero(&isa, sizeof(isa));

    /* Standard header initialization */
    isa.isa_maj = ISAKMP_MAJOR_VERSION;
    isa.isa_min = ISAKMP_MINOR_VERSION;
    isa.isa_np = ISAKMP_NEXT_SA;
    isa.isa_xchg = ISAKMP_XCHG_IDPROT;
    isa.isa_length = sizeof(isa);
    /* R-cookie, flags and MessageID is left zero */

    /* Initiator cookie */
    get_cookie(ISAKMP_INITIATOR, isa.isa_icookie, COOKIE_SIZE, (struct sockaddr *)&sin);

    st = get_state();
    st->st_ouruser = sin2;
    st->st_ournetmask = sin3;
    st->st_peeruser = sin4;
    st->st_peernetmask = sin5;

    bcopy(&sin, &(st->st_peer), sizeof(sin));

    /* Get an ISAKMP proposal (Oakley) */
    p = oakley_get_sa(&j, sin, GOAL_KEYMANAGEMENT, st);
    if (p == (u_char *) NULL)
    {
	free_state(st);
	return;
    }

    isa.isa_length += j;
    isa.isa_length = htonl(isa.isa_length);

    buffer = (u_char *) calloc(sizeof(isa) + j, sizeof(u_char));
    if (buffer == (u_char *) NULL)
      exit_log("calloc() failed in oakley_initiate()");

    bcopy(&isa, buffer, sizeof(isa));
    bcopy(p, buffer + sizeof(isa), j);

    isasa = (struct isakmp_sa *) (buffer + sizeof(isa));
    st->st_situation = SIT_IDENTITY_ONLY;
    st->st_goal = goal;
    st->st_state = OAKLEY_MAIN_I_1;
    bcopy(isa.isa_icookie, st->st_icookie, COOKIE_SIZE);
    st->st_packet = buffer;
    st->st_packet_len = sizeof(isa) + j;
    st->st_doi = ISAKMP_DOI_IPSEC;

    /* Copy sent SA */
    st->st_sa = (u_char *) calloc(j, sizeof(u_char));
    if (st->st_sa == (u_char *) NULL)
      exit_log("calloc() failed in oakley_initiate()");
    bcopy(p, st->st_sa, j);
    free(p);

    insert_state(st);

    /* Transmit */
    if (sendto(sock, buffer, sizeof(isa) + j, 0, (struct sockaddr *)&sin,
	       sizeof(sin)) != sizeof(isa) + j)
      log(1, "sendto() failed in oakley_initiate()");
#ifdef DEBUG
    else
      log(0, "transmitted %d bytes", sizeof(isa) + j);
#endif

    /* Set up a retransmission event, half a minute henceforth */
    event_schedule(EVENT_RETRANSMIT, EVENT_RETRANSMIT_DELAY, st, 0);
}

/*
 * Demux for IPsec DOI initiator.
 */
void
ipsecdoi_initiate(int sock, struct sockaddr_in sin, struct sockaddr_in sin2,
		  struct sockaddr_in sin3, struct sockaddr_in sin4,
		  struct sockaddr_in sin5, int goal, int protocol)
{
    switch (protocol)
    {
	case KEY_OAKLEY:
	    oakley_initiate(sock, sin, sin2, sin3, sin4, sin5, goal);
	    break;

	default:
	    log(0, "unsupported protocol %d specified in ipsecdoi_initiate()",
		protocol);
	    return;
    }
}

/*
 * Examine a list of IPsec DOI policies for IPsec and copy the
 * them into the state.
 * XXX The policy is hardwired now; we only accept a few select
 * XXX algorithms.
 * XXX It also doesn't handle myltiprotocol policies.
 * XXX This should change so that 0 is the value returned if all
 * XXX went fine; otherwise, return a notification error type.
 */
static int
ipsecdoi_ipsec_sa_copy(u_char *from, int length,
		       struct sockaddr sa, struct state *st)
{
    struct isakmp_proposal *isapf;
    struct isakmp_attribute *isaat;
    struct isakmp_transform *isat;
    struct isakmp_sa *isasa;
    struct isakmp_generic *isag;
    int l, k, j, i;
    int acceptable, len, d;
    int saduration, salifetype, saduration2;
    int groupdescription, encapsulate, hmacalg;
    int protoid, transid;

    isag = (struct isakmp_generic *) (from + sizeof(struct isakmp_hdr));
    i = sizeof(struct isakmp_hdr) + sizeof(struct isakmp_sa) +
	IPSEC_DOI_SITUATION_LENGTH + ntohs(isag->isag_length);

    isasa = (struct isakmp_sa *) (from + sizeof(struct isakmp_hdr) +
				  ntohs(isag->isag_length));
    len = ntohs(isasa->isasa_length) - sizeof(struct isakmp_sa) -
	  IPSEC_DOI_SITUATION_LENGTH;

    /* Verify that DOI is IPsec */
    if (ntohl(isasa->isasa_doi) != ISAKMP_DOI_IPSEC)
#ifdef DEBUG
    {
        log(0, "unsupported DOI %d from %s, port %d", ntohl(isasa->isasa_doi),
	    get_address(st->st_peer), get_port(st->st_peer));
#endif
        return 0;
#ifdef DEBUG
    }
#endif

    /* Verify that this is a supported situation */
    bcopy(from + sizeof(struct isakmp_hdr) + sizeof(struct isakmp_sa) +
	  ntohs(isag->isag_length), &d, sizeof(d));
    if (ntohl(d) != SIT_IDENTITY_ONLY)
#ifdef DEBUG
    {
        log(0, "unsupported situation %d from %s, port %d", d,
	    get_address(st->st_peer), get_port(st->st_peer));
#endif
        return 0;
#ifdef DEBUG
    }
#endif

    while (1)
    {
	/* Set up some defaults */
	saduration = SA_LIFE_DURATION_DEFAULT;
	saduration2 = 0;
	encapsulate = ENCAPSULATION_MODE_TUNNEL;
	salifetype = 0;
	groupdescription = 0;
	hmacalg = 0;

	isapf = (struct isakmp_proposal *) (from + i);
	acceptable = 1;

	/* Length mismatch */
	if (ntohs(isapf->isap_length) > len)
	  return 0;

	j = i + sizeof(struct isakmp_proposal) + isapf->isap_spisize;
	isat = (struct isakmp_transform *) (from + j);

	if (ntohs(isat->isat_length) > len + sizeof(struct isakmp_proposal) +
	    isapf->isap_spisize)
	  return 0;

	protoid = isapf->isap_protoid;
	transid = isat->isat_transid;

#ifdef DEBUG
	log(0, "protocol in proposal %d, transform %d", protoid, transid);
#endif

	/* Check that the transform IDs/Protocols are valid */
	switch (protoid)
	{
	    case PROTO_IPSEC_AH:
		if ((transid != AH_MD5_KPDK) &&
		    (transid != AH_MD5) &&
		    (transid != AH_SHA))
#ifdef DEBUG
		{
		    log(0, "unknown/unsupported transform %d in AH proposal",
			transid);
#endif
		    return 0;
#ifdef DEBUG
		}
#endif
		break;

	    case PROTO_IPSEC_ESP:
		if ((transid != ESP_DES_IV64) &&
		    (transid != ESP_DES) &&
		    (transid != ESP_DES_IV32) &&
		    (transid != ESP_3DES) &&
		    (transid != ESP_RC5))
#ifdef DEBUG
		{
		    log(0, "unknown/unsupported transform %d in AH proposal",
			transid);
#endif
		    return 0;
#ifdef DEBUG
		}
#endif
		break;

	    default:
#ifdef DEBUG
		log(0, "unknown/unsupported protocol id %d", protoid);
#endif
		return 0;
	}

	if ((isapf->isap_spisize != IPSEC_DOI_SPI_SIZE) ||
	    (isapf->isap_notrans != 1) ||
	    (isat->isat_np != ISAKMP_NEXT_NONE))
	{
	    /* Next proposal, if there */
	    if (isapf->isap_np == ISAKMP_NEXT_P)
	    {
		i += ntohs(isapf->isap_length);

		if (length - i <= sizeof(struct isakmp_proposal))
		  return 0;

		len += ntohs(isapf->isap_length) + isapf->isap_spisize;
		isapf = (struct isakmp_proposal *) (from + i);

		if (ntohs(isapf->isap_length) + i > length)
		  return 0;

		continue;
	    }
	    else
	      return 0;
	}

	j += sizeof(struct isakmp_transform);

	k = ntohs(isat->isat_length) - sizeof(struct isakmp_transform);
	while (k != 0)
	{
	    isaat = (struct isakmp_attribute *) (from + j);
	    isaat->isaat_type = ntohs(isaat->isaat_type);
	    l = ntohs(isaat->isaat_lv);

	    if (isaat->isaat_af)
	    {
		j += sizeof(struct isakmp_attribute);
		k -= sizeof(struct isakmp_attribute);
	    }
	    else
	    {
		j += l;
		k -= l;
	    }

	    /*
	     * XXX We should also be checking whether the attribute
	     * is/should be TLV or TV
	     */
	    switch (isaat->isaat_rtype)
	    {
		case SA_LIFE_TYPE:
#ifdef DEBUG
		    log(0, "SA life type %d", l);
#endif
		    if ((l != SA_LIFE_TYPE_SECONDS) &&
			(l != SA_LIFE_TYPE_KBYTES))
		      acceptable = 0;
		    else
		      salifetype = l;

		    break;

		case SA_LIFE_DURATION:
#ifdef DEBUG
		    log(0, "SA life duration %u", l);
#endif
		    switch (salifetype)
		    {
			case SA_LIFE_TYPE_SECONDS:
			    if ((l <= SA_LIFE_DURATION_MINIMUM) ||
				(l >= SA_LIFE_DURATION_MAXIMUM))
			      acceptable = 0;
			    else
			      saduration = l;

			    break;

			case SA_LIFE_TYPE_KBYTES:
			    if (l <= SA_LIFE_DURATION_K_MINIMUM)
			      acceptable = 0;
			    else
			      saduration2 = l;

			    break;

			default:
#ifdef DEBUG
			    log(0, "life duration given without lifetype");
#endif
			    acceptable = 0;
		    }

		    break;

		case GROUP_DESCRIPTION:
#ifdef DEBUG
		    log(0, "group description %d", l);
#endif
		    if (l != OAKLEY_DEFAULT_GROUP)
		      acceptable = 0;
		    else
		      groupdescription = l;
		    break;

		case AUTH_ALGORITHM:
#ifdef DEBUG
		    log(0, "AUTH algorithm %d", l);
#endif
		    if ((l != AUTH_ALGORITHM_HMAC_MD5) &&
			(l != AUTH_ALGORITHM_HMAC_SHA1))
		      acceptable = 0;
		    else
		      hmacalg = l;
		    break;

		case ENCAPSULATION_MODE:
#ifdef DEBUG
		    log(0, "encapsulation mode %d", l);
#endif
		    if ((l != ENCAPSULATION_MODE_TUNNEL) &&
			(l != ENCAPSULATION_MODE_TRANSPORT))
		      acceptable = 0;
		    else
		      encapsulate = l;
		    break;

		default:
		    acceptable = 0;
		    log(0, "unsupported attribute %d from %s, port %d",
			isaat->isaat_rtype, get_address(sa), get_port(sa));
		    break;
	    }

	    isaat->isaat_type = htons(isaat->isaat_type);

	    /* Something unacceptable with this policy */
	    if (acceptable == 0)
	      break;
	}

	/*
	 * Something went wrong.
	 */
	if (acceptable == 0)
	{
	    /* Next proposal, if there */
	    if (isapf->isap_np == ISAKMP_NEXT_P)
	    {
		i += ntohs(isapf->isap_length);

		if (length - i <= sizeof(struct isakmp_proposal))
		  return 0;

		len += ntohs(isapf->isap_length) + isapf->isap_spisize;
		isapf = (struct isakmp_proposal *) (from + i);

		if (ntohs(isapf->isap_length) + i > length)
		  return 0;

		continue;
	    }
	    else
	      return 0;
	}

#if 0   /* My mistake, this code should never have been written */
	/* Copy proposal */
	bcopy(isapf, isapt, ntohs(isapf->isap_length));

	isasa->isasa_length = htons(ntohs(isapf->isap_length) +
				    sizeof(struct isakmp_sa) +
				    IPSEC_DOI_SITUATION_LENGTH);

	st->st_proposal = (u_char *) calloc(ntohs(isapt->isap_length),
					    sizeof(u_char));
	if (st->st_proposal == (u_char *) NULL)
	  exit_log("calloc() failed in ipsecdoi_oakley_sa_copy()");

	bcopy(isapt, st->st_proposal, ntohs(isapt->isap_length));
#endif

	/* Copy SPI */
	st->st_peer_spi = (u_char *) calloc(IPSEC_DOI_SPI_SIZE,
					    sizeof(u_char));
	if (st->st_peer_spi == (u_char *) NULL)
	  exit_log("calloc() failed in ipsecdoi_oakleu_sa_copy()");

	bcopy((u_char *)isapf + sizeof(struct isakmp_proposal),
	      st->st_peer_spi, IPSEC_DOI_SPI_SIZE);
	st->st_peer_spi_len = IPSEC_DOI_SPI_SIZE;

	/* Copy all accepted attributes */
	st->st_peer_expire = saduration;
	st->st_peer_sa_duration = saduration2;
	st->st_peer_group = groupdescription;
	st->st_peer_encapsulation_mode = encapsulate;
	st->st_peer_hmac_algorithm = hmacalg;
	st->st_peer_protoid = protoid;
	st->st_peer_transid = transid;

#ifdef DEBUG
	fprintf(stdout, "SPI accepted (%d): ", st->st_peer_spi_len);
	for (l = 0; l < st->st_peer_spi_len; l++)
	  fprintf(stdout, "%02x ", st->st_peer_spi[l]);
	fprintf(stdout, "\n");
	fprintf(stdout, "accepted protocol %d, transform %d\n", protoid,
		transid);
	fprintf(stdout, "SA expiration %u seconds, %u kilobytes\n", saduration,
	    saduration2);
	fprintf(stdout, "encapsulation mode %d, AUTH algorithm %d\n",
	    encapsulate, hmacalg);
	fprintf(stdout, "group description %d\n", groupdescription);
#endif

	return 1;
    }
}

/*
 * Examine a list of IPsec DOI policies and copy the acceptable one. If
 * none is acceptable, return 0. Also copy the accepted proposal in the
 * state object.
 * XXX The policy is hardwired now; we only accept a policy
 * XXX with the Oakley transform.
 * XXX It also doesn't handle policies that span multiple protocols.
 */
static int
ipsecdoi_oakley_sa_copy(u_char *from, u_char *to, int length,
			struct sockaddr sa, struct state *st)
{
    struct isakmp_proposal *isapf, *isapt;
    struct isakmp_attribute *isaat;
    struct isakmp_transform *isat;
    struct isakmp_sa *isasa;
    int l, k, j, i = sizeof(struct isakmp_hdr) + sizeof(struct isakmp_sa) +
		     IPSEC_DOI_SITUATION_LENGTH;
    int enc, hash, group, auth;
    int acceptable, len, d;
    u_char *pp;

    isapt = (struct isakmp_proposal *) (to + i);

    isasa = (struct isakmp_sa *) (from + sizeof(struct isakmp_hdr));
    len = ntohs(isasa->isasa_length) - sizeof(struct isakmp_sa) -
	  IPSEC_DOI_SITUATION_LENGTH;

    isasa = (struct isakmp_sa *) (to + sizeof(struct isakmp_hdr));

    while (1)
    {
	enc = 0;
	hash = 0;
	group = 0;
	auth = 0;
	isapf = (struct isakmp_proposal *) (from + i);
	acceptable = 1;

	/* Length mismatch */
	if (ntohs(isapf->isap_length) > len)
	  return 0;

	j = i + sizeof(struct isakmp_proposal) + isapf->isap_spisize;
	isat = (struct isakmp_transform *) (from + j);

	if (ntohs(isat->isat_length) > len + sizeof(struct isakmp_proposal) +
	    isapf->isap_spisize)
	  return 0;

	if ((isapf->isap_protoid != PROTO_ISAKMP) ||
	    ((isapf->isap_spisize != IPSEC_DOI_SPI_SIZE) &&
	     (isapf->isap_spisize != 0)) ||
	    (isapf->isap_notrans != 1) ||
	    (isat->isat_np != ISAKMP_NEXT_NONE) ||
	    (isat->isat_transid != KEY_OAKLEY))
	{
	    /* Next proposal, if there */
	    if (isapf->isap_np == ISAKMP_NEXT_P)
	    {
		i += ntohs(isapf->isap_length);

		if (length - i <= sizeof(struct isakmp_proposal))
		  return 0;

		len += ntohs(isapf->isap_length) + isapf->isap_spisize;
		isapf = (struct isakmp_proposal *) (from + i);

		if (ntohs(isapf->isap_length) + i > length)
		  return 0;

		continue;
	    }
	    else
	      return 0;
	}

	st->st_transid = isat->isat_transid;
	st->st_protoid = isapf->isap_protoid;

	j += sizeof(struct isakmp_transform);

	k = ntohs(isat->isat_length) - sizeof(struct isakmp_transform);
	while (k != 0)
	{
	    isaat = (struct isakmp_attribute *) (from + j);
	    isaat->isaat_type = ntohs(isaat->isaat_type);

	    l = ntohs(isaat->isaat_lv);

	    if (isaat->isaat_af)
	    {
		j += sizeof(struct isakmp_attribute);
		k -= sizeof(struct isakmp_attribute);
	    }
	    else
	    {
		j += l;
		k -= l;
	    }

	    /*
	     * XXX We should also be checking whether the attribute
	     * is/should be TLV or TV
	     */
	    switch (isaat->isaat_rtype)
	    {
		case OAKLEY_ENCRYPTION_ALGORITHM:
		    enc = 1;
#ifdef DEBUG
		    log(0, "encryption algorithm %d", l);
#endif
		    if (l != OAKLEY_DES_CBC)
		      acceptable = 0;
		    st->st_enc = l;
		    break;

		case OAKLEY_HASH_ALGORITHM:
#ifdef DEBUG
		    log(0, "hash algorithm %d", l);
#endif
		    if ((l != OAKLEY_MD5) &&
			(l != OAKLEY_SHA))
		      acceptable = 0;
		    st->st_hash = l;
		    hash = 1;
		    break;

		case OAKLEY_AUTHENTICATION_METHOD:
#ifdef DEBUG
		    log(0, "authentication method %d", l);
#endif
		    if ((l != OAKLEY_PRESHARED_KEY) &&
			(l != OAKLEY_RSA_SIG))
		      acceptable = 0;

		    if (l == OAKLEY_PRESHARED_KEY)
		    {
			pp = get_preshared_key(st, &d);
			if (pp == (u_char *) NULL)
			  acceptable = 0;
			else
			  free(pp);

#ifdef DEBUG
			if (acceptable == 0)
			  log(0, "get a preshared key with %s first!",
			      get_address(sa));
#endif
		    }

		    st->st_auth = l;
		    auth = 1;
		    break;

		case OAKLEY_GROUP_DESCRIPTION:
#ifdef DEBUG
		    log(0, "group description %d", l);
#endif
		    if (l != OAKLEY_DEFAULT_GROUP)
		      acceptable = 0;
		    st->st_group = l;
		    group = 1;
		    break;

		case OAKLEY_LIFE_TYPE:
#ifdef DEBUG
		    log(0, "life type %d", l);
#endif
		    if (l != OAKLEY_LIFE_SECONDS)
		      acceptable = 0;
		    break;

		case OAKLEY_LIFE_DURATION:
#ifdef DEBUG
		    log(0, "life duration %d", l);
#endif
		    if ((l <= OAKLEY_ISAKMP_SA_LIFETIME_MINIMUM) ||
			(l >= OAKLEY_ISAKMP_SA_LIFETIME_MAXIMUM))
		      acceptable = 0;
		    else
		      st->st_expire = l;
		    break;

		    /* XXX Handle more */
		default:
		    acceptable = 0;
		    log(0, "unsupported attribute %d from %s, port %d",
			isaat->isaat_rtype, get_address(sa), get_port(sa));
		    break;
	    }

	    isaat->isaat_type = htons(isaat->isaat_type);

	    /* Something unacceptable with this policy */
	    if (acceptable == 0)
	      break;
	}

	/*
	 * One of the default attributes not negotiated or unacceptable
	 * policy
	 */
	if ((enc == 0) || (auth == 0) || (hash == 0) || (group == 0) ||
	    (acceptable == 0))
	{
	    /* Next proposal, if there */
	    if (isapf->isap_np == ISAKMP_NEXT_P)
	    {
		i += ntohs(isapf->isap_length);

		if (length - i <= sizeof(struct isakmp_proposal))
		  return 0;

		len += ntohs(isapf->isap_length) + isapf->isap_spisize;
		isapf = (struct isakmp_proposal *) (from + i);

		if (ntohs(isapf->isap_length) + i > length)
		  return 0;

		continue;
	    }
	    else
	      return 0;
	}

	/* Copy proposal */
	bcopy(isapf, isapt, ntohs(isapf->isap_length));
	isapt->isap_np = ISAKMP_NEXT_NONE;

	isasa->isasa_length = htons(ntohs(isapf->isap_length) +
				    sizeof(struct isakmp_sa) +
				    IPSEC_DOI_SITUATION_LENGTH);

	st->st_proposal = (u_char *) calloc(ntohs(isapt->isap_length),
					    sizeof(u_char));
	if (st->st_proposal == (u_char *) NULL)
	  exit_log("calloc() failed in ipsecdoi_oakley_sa_copy()");

	bcopy(isapt, st->st_proposal, ntohs(isapt->isap_length));

	/* Copy SPI */
	st->st_spi = (u_char *) calloc(IPSEC_DOI_SPI_SIZE, sizeof(u_char));
	if (st->st_spi == (u_char *) NULL)
	  exit_log("calloc() failed in ipsecdoi_oakleu_sa_copy()");

	bcopy((u_char *)isapf + sizeof(struct isakmp_proposal), st->st_spi,
	      isapf->isap_spisize);
	st->st_spi_len = isapf->isap_spisize;

#ifdef DEBUG
	log(0, "SPI accepted: ");
	for (l = 0; l < st->st_spi_len; l++)
	  fprintf(stderr, "%02x ", st->st_spi[l]);
	fprintf(stderr, "\n");
#endif

	return 1;
    }
}

/*
 * Handle a Main Mode Oakley first packet (responder side).
 */
void
ipsecdoi_handle_rfirst(int sock, u_char *buffer, int length,
		      struct sockaddr sa)
{
    struct isakmp_hdr *isa;
    struct isakmp_sa *isasa, *isas;
    struct state *st;
    u_char *packet;
    int i, k;
    struct sockaddr_in sin;

    isa = (struct isakmp_hdr *) buffer;
    isasa = (struct isakmp_sa *) (buffer + sizeof(struct isakmp_hdr));

    if (ntohs(isasa->isasa_length) + sizeof(struct isakmp_hdr) >
	length)
    {
	free(buffer);
	log(0, "malformed packet from %s, port %d",
	    get_address(sa), get_port(sa));
	send_notification(sock, PAYLOAD_MALFORMED, NULL, 0,
			  PROTO_ISAKMP, isa->isa_icookie, isa->isa_rcookie, 0,
			  sa);
	return;
    }

    /* XXX Check various fields ? */

    /* Set up state */
    st = get_state();

    bcopy(isa->isa_icookie, st->st_icookie, COOKIE_SIZE);

    get_cookie(ISAKMP_RESPONDER, st->st_rcookie, COOKIE_SIZE, &sa);

    st->st_peer = sa;
    st->st_doi = ISAKMP_DOI_IPSEC;
    st->st_situation = SIT_IDENTITY_ONLY; /* We only support this */
    st->st_state = OAKLEY_MAIN_R_1;

    /* The reply packet can't be larger than the original request */
    packet = (u_char *) calloc(length, sizeof(u_char));
    if (packet == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_rfirst()");

    isa = (struct isakmp_hdr *) packet;

    /* Copy header as received, and responder cookie */
    bcopy(buffer, packet, sizeof(struct isakmp_hdr));
    bcopy(st->st_rcookie, isa->isa_rcookie, COOKIE_SIZE);
    isa->isa_length = 0;

    isas = (struct isakmp_sa *) (packet + sizeof(struct isakmp_hdr));

    isas->isasa_doi = ntohl(ISAKMP_DOI_IPSEC);
    i = sizeof(struct isakmp_hdr) + sizeof(struct isakmp_sa) +
	IPSEC_DOI_SITUATION_LENGTH;
    packet[i - 1] = SIT_IDENTITY_ONLY;

    if (ipsecdoi_oakley_sa_copy(buffer, packet, length, sa, st) == 0)
    {
	free(buffer);
	free(packet);
	free_state(st);
	log(0, "no acceptable proposal from %s, port %d",
	    get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    st->st_sa = (u_char *) calloc(ntohs(isasa->isasa_length), sizeof(u_char));
    if (st->st_sa == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_rfirst()");

    bcopy(isas, st->st_sa, ntohs(isasa->isasa_length));

    /* XXX Handle Aggressive Mode here ? */

    free(buffer);

    st->st_packet = packet;

    /* Get my identity */
    k = sizeof(sin);
    if (getsockname(sock, (struct sockaddr *)&sin, &k) < 0)
      exit_log("getsockname() in ipsecdoi_handle_rfirst()");

    if (k <= 0)
      exit_log("getsockname() in ipsecdoi_handle_rfirst()");

    st->st_myidentity = (u_char *) calloc(1, sizeof(struct in_addr));
    if (st->st_myidentity == (u_char *) NULL)
      exit_log("strdup() in ipsecdoi_handle_rfirst()");
    bcopy(&(sin.sin_addr), st->st_myidentity, sizeof(struct in_addr));
    st->st_myidentity_len = sizeof(struct in_addr);
    st->st_myidentity_type = ID_IPV4_ADDR;

    length = ntohs(isas->isasa_length) + sizeof(struct isakmp_hdr);
#ifdef DEBUG
    log(0, "my identity is %s", inet_ntoa(sin.sin_addr));
    log(0, "sending %d bytes to %s, port %d", length,
	get_address(sa), get_port(sa));
#endif
    isa->isa_length = ntohl(length);
    st->st_packet_len = length;

    if (sendto(sock, packet, length, 0, &sa, sizeof(sa)) != length)
      log(1, "sendto() failed in ipsecdoi_handle_rfirst() for %s, port %d",
	  get_address(sa), get_port(sa));

    insert_state(st);
    event_schedule(EVENT_CLEANUP, EVENT_CLEANUP_DELAY, st, 0);
}

/*
 * Get preshared key for remote IP from the SHARED_SECRETS_FILE.
 *
 * Leading and trailing blanks on a line are ignored.
 * Empty lines and those starting with # are treated as comments.
 * Each entry has an IP address and a secret, normally separated by whitespace.
 * The secret can have quotes at either end to indicate that the enclosed
 * Whitespace is significant.
 * If nothing follows the IP address, the whole next line is taken as the
 * secret, including any whitespace and quotes.
 *
 * A sample secret file might be:
 *
 * # sssh! secrets!  Make this file only readable by owner.
 * 19.98.01.24 no way to know
 * 19.98.01.25 " leading and following space "
 * 19.98.01.26
 *	leading and following space
 */
static u_char *
get_preshared_key(struct state *st, int *length)
{
    u_char buffer[513];
    u_char *p = (u_char *) NULL;
    FILE *fp;

#ifdef DEBUG
    log(0, "opening %s", SHARED_SECRETS_FILE);
#endif

    fp = fopen(SHARED_SECRETS_FILE, "r");
    if (fp == (FILE *) NULL)
      exit_log("fopen() failed in get_preshared_key()");

    while (fgets(buffer, sizeof(buffer), fp) != (char *) NULL)
    {
	struct in_addr ia;
	u_char *s, *b, *e;

#ifdef DEBUG
	log(0, "loop while searching for preshared key");
#endif

	/* strip leading and trailing whitespace, including \n */
	for (e = buffer+strlen(buffer); e!=buffer && isspace(e[-1]); e--)
	    ;
	*e = '\0';

	for (s = buffer; isspace(*s); s++)
	    ;
	/* treat empty line or one starting with # as a comment
	 * Note that the secret may start with a # or be empty.
	 */
	if (s[0] == '\0' || s[0] == '#')
	    continue;

	/* find end of first field, the IP address */
	for (b = s; *b != '\0' && !isspace(*b); b++)
	    ;
	if (*b != '\0') {
	    *b = '\0';  /* terminate field */
	    do b++; while (*b != '\0' && isspace(*b));  /* move to next field */
	}

	if (inet_aton(s, &ia) == 0)
	    exit_log("Expecting IP address in %s, but found \"%s\"",
		SHARED_SECRETS_FILE, buffer);

	if (*b == '\0') {
	    /* if no secret field is found on this line, use next line, raw */
	    if (fgets(buffer, sizeof(buffer), fp) == (char *) NULL)
		exit_log("bad secrets file");
	    b = buffer;
	    e = b + strlen(e);
	    if (e != b && e[-1] == '\n')
		*--e = '\0';
	} else {
	    /* strip any quote from either end of secret */
	    if (*b == '"')
		b++;
	    if (e != b && e[-1] == '"')
		*--e = '\0';
	}

	if (st->st_peer.sa_family == AF_INET
	&& bcmp(&ia, &(*(struct sockaddr_in *)&st->st_peer).sin_addr, sizeof(ia)) == 0)
	{
	    p = strdup(b);
	    if (p == (u_char *) NULL)
	      exit_log("strdup() failed in get_preshared_key()");
	    *length = e - b;

#ifdef DEBUG
	    log(0, "secret used is [%s], length = %d", p, strlen(p));
#endif

	    break;
	}
    }
    fclose(fp);
    return p;
}

/*
 * SKEYID for preshared keys.
 */
static void
skeyid_preshared(struct state *st, u_char *id, int length)
{
    int k;
    MD5_CTX md5ctx;
    SHA1_CTX sha1ctx;
    char buffer[HMAC_BUFSIZE], buffer2[HMAC_BUFSIZE];
    struct isakmp_generic *isag;

    bzero(buffer, HMAC_BUFSIZE);
    bzero(buffer2, HMAC_BUFSIZE);

    if (st->st_prf == 0)
    {
	/* Prepare the two pads for the HMAC */
	if (length > HMAC_BUFSIZE)
	{
	    switch (st->st_hash)
	    {
		case OAKLEY_MD5:
		    MD5Init(&md5ctx);
		    MD5Update(&md5ctx, id, length);
		    MD5Final(buffer, &md5ctx);
		    bcopy(buffer, buffer2, MD5_DIGEST_SIZE);
		    break;

		case OAKLEY_SHA:
		    SHA1Init(&sha1ctx);
		    SHA1Update(&sha1ctx, id, length);
		    SHA1Final(buffer, &sha1ctx);
		    bcopy(buffer, buffer2, SHA1_DIGEST_SIZE);
		    break;

		default:
		    exit_log("unknown/unsupported hash algorithm %d in skeyid_preshared()", st->st_hash);
	    }
	}
	else
	{
	    bcopy(id, buffer, length);
	    bcopy(id, buffer2, length);
	}

	for (k = 0; k < HMAC_BUFSIZE; k++)
	{
	    buffer[k] ^= HMAC_IPAD;
	    buffer2[k] ^= HMAC_OPAD;
	}

	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		st->st_skeyid = (u_char *) calloc(MD5_DIGEST_SIZE, sizeof(u_char));
		if (st->st_skeyid == (u_char *) NULL)
		  exit_log("calloc() in skeyid_preshared()");
		st->st_skeyid_len = MD5_DIGEST_SIZE;

		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buffer, HMAC_BUFSIZE);

		k = sizeof(struct isakmp_generic);

		isag = (struct isakmp_generic *) st->st_ni;
		MD5Update(&md5ctx, st->st_ni + k,
			  ntohs(isag->isag_length) - k);

		isag = (struct isakmp_generic *) st->st_nr;
		MD5Update(&md5ctx, st->st_nr + k,
			  ntohs(isag->isag_length) - k);

		MD5Final(st->st_skeyid, &md5ctx);
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buffer2, HMAC_BUFSIZE);
		MD5Update(&md5ctx, st->st_skeyid, MD5_DIGEST_SIZE);
		MD5Final(st->st_skeyid, &md5ctx);
		return;

	    case OAKLEY_SHA:
		st->st_skeyid = (u_char *) calloc(SHA1_DIGEST_SIZE, sizeof(u_char));
		if (st->st_skeyid == (u_char *) NULL)
		  exit_log("calloc() in skeyid_preshared()");
		st->st_skeyid_len = SHA1_DIGEST_SIZE;

		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buffer, HMAC_BUFSIZE);

		k = sizeof(struct isakmp_generic);

		isag = (struct isakmp_generic *) st->st_ni;
		SHA1Update(&sha1ctx, st->st_ni + k,
			  ntohs(isag->isag_length) - k);

		isag = (struct isakmp_generic *) st->st_nr;
		SHA1Update(&sha1ctx, st->st_nr + k,
			  ntohs(isag->isag_length) - k);

		SHA1Final(st->st_skeyid, &sha1ctx);
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buffer2, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, st->st_skeyid, SHA1_DIGEST_SIZE);
		SHA1Final(st->st_skeyid, &sha1ctx);
		return;

	    default:
		exit_log("unknown hash algorithm %d", st->st_hash);
	}
    }
    else
    {
	/* XXX Handle 3DES */
    }
}


/*
 * Generate the SKEYID_*
 */
static void
generate_skeyids(struct state *st)
{
    u_char *id;
    int len, k;
    MD5_CTX md5ctx;
    SHA1_CTX sha1ctx;
    u_char buf1[HMAC_BUFSIZE], buf2[HMAC_BUFSIZE];
    u_char *p, c;
    MP_INT temp, temp2;
    u_char *gi, *gr;
    int gilen, grlen;

    mpz_init(&temp);
    mpz_init(&temp2);

    bzero(buf1, HMAC_BUFSIZE);
    bzero(buf2, HMAC_BUFSIZE);

    gilen = (mpz_sizeinbase(&(st->st_gi), 16) + 1) / 2;
    gi = (u_char *) calloc(gilen, sizeof(u_char));
    if (gi == (u_char *) NULL)
      exit_log("calloc() failed in get_hash()");

    mpz_set(&temp, &(st->st_gi));

    for (k = gilen - 1 ; k >= 0; k--)
    {
	gi[k] = mpz_mdivmod_ui(&temp2, NULL, &temp, 256);
	mpz_set(&temp, &temp2);
    }

    grlen = (mpz_sizeinbase(&(st->st_gr), 16) + 1) / 2;
    gr = (u_char *) calloc(grlen, sizeof(u_char));
    if (gr == (u_char *) NULL)
      exit_log("calloc() failed in get_hash()");

    mpz_set(&temp, &(st->st_gr));

    for (k = grlen - 1; k >= 0; k--)
    {
	gr[k] = mpz_mdivmod_ui(&temp2, NULL, &temp, 256);
	mpz_set(&temp, &temp2);
    }

    mpz_clear(&temp);
    mpz_clear(&temp2);

    switch (st->st_auth)
    {
	case OAKLEY_PRESHARED_KEY:
	    /* Generate the SKEYID */
	    id = get_preshared_key(st, &len);
	    skeyid_preshared(st, id, len);
	    free(id);
	    break;

	case OAKLEY_DSS_SIG:
	case OAKLEY_RSA_SIG:
	    /* XXX */
	    break;

	case OAKLEY_RSA_ENC:
	    /* XXX */
	    break;

	default:
	    exit_log("generate_skeyids(): unknown authentication method");
    }

    /* Get the g^xy for HMAC'ing */
    len = (mpz_sizeinbase(&(st->st_shared), 16) + 1) / 2;

#ifdef DEBUG
    log(0, "size of g^xy is %d", len);
#endif

    p = (u_char *) calloc(len, sizeof(u_char));
    if (p == (u_char *) NULL)
      exit_log("calloc() failed in generate_skeyids()");

    mpz_init(&temp);
    mpz_init(&temp2);
    mpz_set(&temp, &(st->st_shared));

    for (k = len - 1; k >= 0; k--)
    {
	p[k] = mpz_mdivmod_ui(&temp2, NULL, &temp, 256);
	mpz_set(&temp, &temp2);
    }

    mpz_clear(&temp);
    mpz_clear(&temp2);

    if (st->st_prf == 0)
    {
	/* Prepare the two pads for the HMAC */
	if (st->st_skeyid_len <= HMAC_BUFSIZE)
	{
	    bcopy(st->st_skeyid, buf1, st->st_skeyid_len);
	    bcopy(st->st_skeyid, buf2, st->st_skeyid_len);
	}
	else
	{
	    switch (st->st_hash)
	    {
		case OAKLEY_MD5:
		    MD5Init(&md5ctx);
		    MD5Update(&md5ctx, st->st_skeyid, st->st_skeyid_len);
		    MD5Final(buf1, &md5ctx);
		    bcopy(buf1, buf2, MD5_DIGEST_SIZE);
		    break;

		case OAKLEY_SHA:
		    SHA1Init(&sha1ctx);
		    SHA1Update(&sha1ctx, st->st_skeyid, st->st_skeyid_len);
		    SHA1Final(buf1, &sha1ctx);
		    bcopy(buf1, buf2, SHA1_DIGEST_SIZE);
		    break;

		default:
		    exit_log("unknown/unsupported hash algorithm %d in generate_skeyids()", st->st_hash);
	    }
	}

	for (k = 0; k < HMAC_BUFSIZE; k++)
	{
	    buf1[k] ^= HMAC_IPAD;
	    buf2[k] ^= HMAC_OPAD;
	}

	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		st->st_skeyid_d_len = MD5_DIGEST_SIZE;
		st->st_skeyid_d = (u_char *) calloc(MD5_DIGEST_SIZE, sizeof(u_char));
		if (st->st_skeyid_d == (u_char *) NULL)
		  exit_log("calloc() in generate_skeyids()");

		st->st_skeyid_a_len = MD5_DIGEST_SIZE;
		st->st_skeyid_a = (u_char *) calloc(MD5_DIGEST_SIZE, sizeof(u_char));
		if (st->st_skeyid_a == (u_char *) NULL)
		  exit_log("calloc() in generate_skeyids()");

		st->st_skeyid_e_len = MD5_DIGEST_SIZE;
		st->st_skeyid_e = (u_char *) calloc(MD5_DIGEST_SIZE, sizeof(u_char));
		if (st->st_skeyid_e == (u_char *) NULL)
		  exit_log("calloc() in generate_skeyids()");

		/* SKEYID_D */
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf1, HMAC_BUFSIZE);
		MD5Update(&md5ctx, p, len);
		MD5Update(&md5ctx, st->st_icookie, COOKIE_SIZE);
		MD5Update(&md5ctx, st->st_rcookie, COOKIE_SIZE);
		c = 0;
		MD5Update(&md5ctx, &c, 1);
		MD5Final(st->st_skeyid_d, &md5ctx);
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf2, HMAC_BUFSIZE);
		MD5Update(&md5ctx, st->st_skeyid_d, st->st_skeyid_d_len);
		MD5Final(st->st_skeyid_d, &md5ctx);

		/* SKEYID_A */
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf1, HMAC_BUFSIZE);
		MD5Update(&md5ctx, st->st_skeyid_d, st->st_skeyid_d_len);
		MD5Update(&md5ctx, p, len);
		MD5Update(&md5ctx, st->st_icookie, COOKIE_SIZE);
		MD5Update(&md5ctx, st->st_rcookie, COOKIE_SIZE);
		c = 1;
		MD5Update(&md5ctx, &c, 1);
		MD5Final(st->st_skeyid_a, &md5ctx);
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf2, HMAC_BUFSIZE);
		MD5Update(&md5ctx, st->st_skeyid_a, st->st_skeyid_a_len);
		MD5Final(st->st_skeyid_a, &md5ctx);

		/* SKEYID_E */
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf1, HMAC_BUFSIZE);
		MD5Update(&md5ctx, st->st_skeyid_a, st->st_skeyid_a_len);
		MD5Update(&md5ctx, p, len);
		MD5Update(&md5ctx, st->st_icookie, COOKIE_SIZE);
		MD5Update(&md5ctx, st->st_rcookie, COOKIE_SIZE);
		c = 2;
		MD5Update(&md5ctx, &c, 1);
		MD5Final(st->st_skeyid_e, &md5ctx);
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf2, HMAC_BUFSIZE);
		MD5Update(&md5ctx, st->st_skeyid_e, st->st_skeyid_e_len);
		MD5Final(st->st_skeyid_e, &md5ctx);

		/* IV */
		st->st_iv_len = MD5_DIGEST_SIZE;
		st->st_iv = (u_char *) calloc(st->st_iv_len, sizeof(u_char));
		if (st->st_iv == (u_char *) NULL)
		  exit_log("calloc() failed in generate_skeyids()");

		MD5Init(&md5ctx);
		MD5Update(&md5ctx, gi, gilen);
		MD5Update(&md5ctx, gr, grlen);
		MD5Final(st->st_iv, &md5ctx);

		break;

	    case OAKLEY_SHA:
		st->st_skeyid_d_len = SHA1_DIGEST_SIZE;
		st->st_skeyid_d = (u_char *) calloc(SHA1_DIGEST_SIZE, sizeof(u_char));
		if (st->st_skeyid_d == (u_char *) NULL)
		  exit_log("calloc() failed in generate_skeyids()");

		st->st_skeyid_a_len = SHA1_DIGEST_SIZE;
		st->st_skeyid_a = (u_char *) calloc(SHA1_DIGEST_SIZE, sizeof(u_char));
		if (st->st_skeyid_a == (u_char *) NULL)
		  exit_log("calloc() in generate_skeyids()");

		st->st_skeyid_e_len = SHA1_DIGEST_SIZE;
		st->st_skeyid_e = (u_char *) calloc(SHA1_DIGEST_SIZE, sizeof(u_char));
		if (st->st_skeyid_e == (u_char *) NULL)
		  exit_log("calloc() in generate_skeyids()");

		/* SKEYID_D */
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf1, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, p, len);
		SHA1Update(&sha1ctx, st->st_icookie, COOKIE_SIZE);
		SHA1Update(&sha1ctx, st->st_rcookie, COOKIE_SIZE);
		c = 0;
		SHA1Update(&sha1ctx, &c, 1);
		SHA1Final(st->st_skeyid_d, &sha1ctx);
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf2, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, st->st_skeyid_d, st->st_skeyid_d_len);
		SHA1Final(st->st_skeyid_d, &sha1ctx);

		/* SKEYID_A */
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf1, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, st->st_skeyid_d, st->st_skeyid_d_len);
		SHA1Update(&sha1ctx, p, len);
		SHA1Update(&sha1ctx, st->st_icookie, COOKIE_SIZE);
		SHA1Update(&sha1ctx, st->st_rcookie, COOKIE_SIZE);
		c = 1;
		SHA1Update(&sha1ctx, &c, 1);
		SHA1Final(st->st_skeyid_a, &sha1ctx);
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf2, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, st->st_skeyid_a, st->st_skeyid_a_len);
		SHA1Final(st->st_skeyid_a, &sha1ctx);

		/* SKEYID_E */
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf1, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, st->st_skeyid_a, st->st_skeyid_a_len);
		SHA1Update(&sha1ctx, p, len);
		SHA1Update(&sha1ctx, st->st_icookie, COOKIE_SIZE);
		SHA1Update(&sha1ctx, st->st_rcookie, COOKIE_SIZE);
		c = 2;
		SHA1Update(&sha1ctx, &c, 1);
		SHA1Final(st->st_skeyid_e, &sha1ctx);
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf2, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, st->st_skeyid_e, st->st_skeyid_e_len);
		SHA1Final(st->st_skeyid_e, &sha1ctx);

                /* IV */
		st->st_iv_len = SHA1_DIGEST_SIZE;
		st->st_iv = (u_char *) calloc(st->st_iv_len, sizeof(u_char));
		if (st->st_iv == (u_char *) NULL)
		  exit_log("calloc() failed in generate_skeyids()");

		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, gi, gilen);
		SHA1Update(&sha1ctx, gr, grlen);
		SHA1Final(st->st_iv, &sha1ctx);

		break;

	    default:
		exit_log("generate_skeyids(): unknown hash function");
	}
    }
    else
    {
	/* XXX Handle 3DES */
    }

    free(gi);
    free(gr);
    free(p);
}

/*
 * Handle HDR;SA from responder (verify proposal), and send back
 * a HDR;KE;Ni.
 */
void
ipsecdoi_handle_i1(int sock, u_char *buffer, int length, struct sockaddr sa,
		   struct state *st)
{
    struct sockaddr_in sa2;
    struct isakmp_hdr *isa, *isa2;
    struct isakmp_sa *isasa;
    struct isakmp_proposal *isap;
    struct isakmp_generic *isag;
    u_int32_t situation;
    u_char tmp[LOCALSECRETSIZE];
    u_char *packet;
    MP_INT temp, temp2;
    int i, k;

    isa = (struct isakmp_hdr *) buffer;

    if (length <= sizeof(struct isakmp_hdr) + sizeof(struct isakmp_sa))
    {
	free(buffer);
	log(0, "too short (%d bytes) packet received from %s, port %d", length,
	    get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    isasa = (struct isakmp_sa *) (buffer + sizeof(struct isakmp_hdr));

    if (htons(isasa->isasa_length) + sizeof(struct isakmp_hdr) !=
	length)
    {
	free(buffer);
	log(0, "malformed HDR;SA packet from %s, port %d",
	    get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    i = sizeof(struct isakmp_hdr) + sizeof(struct isakmp_sa);

    /* Not sure if this should ever happen...i'm just paranoid */
    bcopy(buffer + i, &situation, IPSEC_DOI_SITUATION_LENGTH);
    situation = ntohl(situation);
    if (situation != SIT_IDENTITY_ONLY)
    {
	free(buffer);
	log(0, "bad situation %d from %s, port %d", situation,
	    get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    i += IPSEC_DOI_SITUATION_LENGTH;

    isap = (struct isakmp_proposal *) (buffer + i);
    if (ntohs(isap->isap_length) + i != length)
    {
	free(buffer);
	log(0, "bad proposal length from %s, port %d",
	    get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    i += sizeof(struct isakmp_proposal) + isap->isap_spisize;

    /* XXX Verify proposal we received back */

    mpz_init(&(st->st_sec));
    mpz_init(&(st->st_gi));
    st->st_sec_in_use = 1;
    st->st_gi_in_use = 1;

    mpz_init(&temp);
    mpz_init(&temp2);

    get_rnd_bytes(tmp, LOCALSECRETSIZE);

    for (k = 0; k < LOCALSECRETSIZE; k++)
    {
	mpz_mul_ui(&(st->st_gi), &(st->st_sec), 256);
	mpz_add_ui(&(st->st_sec), &(st->st_gi), tmp[k]);
    }

    mpz_powm(&(st->st_gi), &groupgenerator, &(st->st_sec), &groupmodulo);

    i = sizeof(struct isakmp_hdr) + 2 * sizeof(struct isakmp_generic) +
	(1 + mpz_sizeinbase(&(st->st_gi), 16)) / 2 + DEFAULTNONCESIZE;

    packet = (u_char *) calloc(i, sizeof(u_char));
    if (packet == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_i1()");

    bcopy(isa, packet, sizeof(struct isakmp_hdr));

    isa2 = (struct isakmp_hdr *) packet;
    isa2->isa_np = ISAKMP_NEXT_KE;
    isag = (struct isakmp_generic *) (packet + sizeof(struct isakmp_hdr));
    isag->isag_np = ISAKMP_NEXT_NONCE;
    isa2->isa_length = htonl(i);
    isag->isag_length = htons((mpz_sizeinbase(&(st->st_gi), 16) + 1) / 2 +
			      sizeof(struct isakmp_generic));

    i -= (sizeof(struct isakmp_generic) + DEFAULTNONCESIZE);

#ifdef DEBUG
    log(0, "Local secret: ");
    mpz_out_str(stderr, 16, &(st->st_sec));
    fprintf(stderr, "\n");
    log(0, "Public value sent: ");
    mpz_out_str(stderr, 16, &(st->st_gi));
    fprintf(stderr, "\nPublic value is %d bytes long.\n",
	    (mpz_sizeinbase(&(st->st_gi), 16) + 1) / 2);
#endif

    mpz_set(&temp, &(st->st_gi));

    for (k = i - 1;
	 k >= sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic);
	 k--)
    {
	packet[k] = mpz_mdivmod_ui(&temp2, NULL, &temp, 256);
	mpz_set(&temp, &temp2);
    }

#ifdef DEBUG
    log(0, "k = %d, i = %d, isag->isag_length = %d", k, i,
	ntohs(isag->isag_length));

    if (mpz_cmp_ui(&temp, 0) != 0)
    {
	mpz_out_str(stderr, 16, &temp);
	fprintf(stderr, "\n");
	exit_log("ipsecdoi_handle_i1(): remainder not zero");
    }
#endif

    mpz_clear(&temp);
    mpz_clear(&temp2);

    /* Generate Ni */
    isag = (struct isakmp_generic *) (packet + i);
    isag->isag_length = htons(sizeof(struct isakmp_generic) +
			      DEFAULTNONCESIZE);
    get_rnd_bytes(packet + i + sizeof(struct isakmp_generic),
		  DEFAULTNONCESIZE);

    /* Copy the Ni nonce in the state object */
    st->st_ni = (u_char *) calloc(ntohs(isag->isag_length), sizeof(u_char));
    if (st->st_ni == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_i1()");

    bcopy(isag, st->st_ni, ntohs(isag->isag_length));

    /* Copy accepted/received proposal in state object */
    st->st_proposal = (u_char *) calloc(ntohs(isap->isap_length),
					sizeof(u_char));
    if (st->st_proposal == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_i1()");

#ifdef DEBUG
    log(0, "copying %d bytes of proposal into state object",
	ntohs(isap->isap_length));
#endif
    bcopy(isap, st->st_proposal, ntohs(isap->isap_length));

    /* Reinsert the state, using the responder cookie we just received */
    delete_state(st);
    bcopy(isa->isa_rcookie, st->st_rcookie, COOKIE_SIZE);
    insert_state(st);

    /* Delete previous retransmission event */
    delete_event(st);

    /* Free previous packet and place the new one in the state object */
    free(st->st_packet);
    st->st_packet_len = ntohl(isa2->isa_length);
    st->st_packet = packet;
    st->st_state = OAKLEY_MAIN_I_2;

    /* Get my identity */
    k = sizeof(sa2);
    if (getsockname(sock, (struct sockaddr *)&sa2, &k) < 0)
      exit_log("getsockname() in ipsecdoi_handle_i1()");

    if (k <= 0)
      exit_log("getsockname() in ipsecdoi_handle_i1()");

    st->st_myidentity = (u_char *) calloc(1, sizeof(struct in_addr));
    if (st->st_myidentity == (u_char *) NULL)
      exit_log("strdup() in ipsecdoi_handle_i1()");
    bcopy(&(sa2.sin_addr), st->st_myidentity, sizeof(struct in_addr));
    st->st_myidentity_len = sizeof(struct in_addr);
    st->st_myidentity_type = ID_IPV4_ADDR;

#ifdef DEBUG
    log(0, "my identity is %s", inet_ntoa(sa2.sin_addr));
#endif

    free(buffer);

    if (sendto(sock, packet, st->st_packet_len, 0, &sa, sizeof(sa)) !=
	st->st_packet_len)
      log(1, "sendto() failed in ipsecdoi_handle_i1()");
#ifdef DEBUG
    else
      log(0, "transmitted %d bytes", st->st_packet_len);
#endif

    /* Schedule for retransmission */
    event_schedule(EVENT_RETRANSMIT, EVENT_RETRANSMIT_DELAY, st, 0);
}

/*
 * If the third argument is 1, generate HASH_I, if 0 generate HASH_R.
 * If fourth argument is 1, I'm the Initiator.
 */
static u_char *
get_hash(struct state *st, int *length, int hashi, int initiator)
{
    u_char *p, *gi, *gr;
    MD5_CTX md5ctx;
    SHA1_CTX sha1ctx;
    u_char buf1[HMAC_BUFSIZE], buf2[HMAC_BUFSIZE];
    int k, gilen, grlen;
    MP_INT temp, temp2;
    struct identity id;
    struct isakmp_sa *isasa;

    mpz_init(&temp);
    mpz_init(&temp2);

    bzero(buf1, HMAC_BUFSIZE);
    bzero(buf2, HMAC_BUFSIZE);

    gilen = (mpz_sizeinbase(&(st->st_gi), 16) + 1) / 2;
    gi = (u_char *) calloc(gilen, sizeof(u_char));
    if (gi == (u_char *) NULL)
      exit_log("calloc() failed in get_hash()");

    mpz_set(&temp, &(st->st_gi));

    for (k = gilen - 1 ; k >= 0; k--)
    {
	gi[k] = mpz_mdivmod_ui(&temp2, NULL, &temp, 256);
	mpz_set(&temp, &temp2);
    }

    grlen = (mpz_sizeinbase(&(st->st_gr), 16) + 1) / 2;
    gr = (u_char *) calloc(grlen, sizeof(u_char));
    if (gr == (u_char *) NULL)
      exit_log("calloc() failed in get_hash()");

    mpz_set(&temp, &(st->st_gr));

    for (k = grlen - 1; k >= 0; k--)
    {
	gr[k] = mpz_mdivmod_ui(&temp2, NULL, &temp, 256);
	mpz_set(&temp, &temp2);
    }

    mpz_clear(&temp);
    mpz_clear(&temp2);

    isasa = (struct isakmp_sa *) st->st_sa;

    if (st->st_prf == 0)
    {
	/* Prepare the two pads for the HMAC */
	if (st->st_skeyid_len <= HMAC_BUFSIZE)
	{
	    bcopy(st->st_skeyid, buf1, st->st_skeyid_len);
	    bcopy(st->st_skeyid, buf2, st->st_skeyid_len);
	}
	else
	{
	    switch (st->st_hash)
	    {
		case OAKLEY_MD5:
		    MD5Init(&md5ctx);
		    MD5Update(&md5ctx, st->st_skeyid, st->st_skeyid_len);
		    MD5Final(buf1, &md5ctx);
		    bcopy(buf1, buf2, MD5_DIGEST_SIZE);
		    break;

		case OAKLEY_SHA:
		    SHA1Init(&sha1ctx);
		    SHA1Update(&sha1ctx, st->st_skeyid, st->st_skeyid_len);
		    SHA1Final(buf1, &sha1ctx);
		    bcopy(buf1, buf2, SHA1_DIGEST_SIZE);
		    break;

		default:
		    exit_log("unknown/unsupported hash algorithm %d in get_hash()", st->st_hash);
	    }
	}

	for (k = 0; k < HMAC_BUFSIZE; k++)
	{
	    buf1[k] ^= HMAC_IPAD;
	    buf2[k] ^= HMAC_OPAD;
	}

	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		*length = MD5_DIGEST_SIZE;
		p = (u_char *) calloc(MD5_DIGEST_SIZE, sizeof(u_char));
		if (p == (u_char *) NULL)
		  exit_log("calloc() failed in get_hash()");

		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf1, HMAC_BUFSIZE);

		if (hashi)
		{
		    MD5Update(&md5ctx, gi, gilen);
		    MD5Update(&md5ctx, gr, grlen);
		    MD5Update(&md5ctx, st->st_icookie, COOKIE_SIZE);
		    MD5Update(&md5ctx, st->st_rcookie, COOKIE_SIZE);
		}
		else
		{
		    MD5Update(&md5ctx, gr, grlen);
		    MD5Update(&md5ctx, gi, gilen);
		    MD5Update(&md5ctx, st->st_rcookie, COOKIE_SIZE);
		    MD5Update(&md5ctx, st->st_icookie, COOKIE_SIZE);
		}

#ifdef DEBUG
		fprintf(stdout, "hashing %d bytes of SA\n",
		        ntohs(isasa->isasa_length) -
			sizeof(struct isakmp_generic));
#endif

		/* SAp */
		MD5Update(&md5ctx, st->st_sa + sizeof(struct isakmp_generic),
			  ntohs(isasa->isasa_length) -
			  sizeof(struct isakmp_generic));

		bzero(&id, sizeof(id));

		if (hashi)
		{
		    if (initiator)
		      id.id_type = st->st_myidentity_type;
		    else
		    {
		        id.id_type = st->st_peeridentity_type;
			id.id_port = st->st_peeridentity_port;
			id.id_protoid = st->st_peeridentity_protocol;
		    }
		}
		else
		{
		    if (initiator)
	  	    {
		        id.id_type = st->st_peeridentity_type;
			id.id_port = st->st_peeridentity_port;
			id.id_protoid = st->st_peeridentity_protocol;
#ifdef DEBUG
			fprintf(stdout, "Port = %d, protoid = %d\n",
			        st->st_peeridentity_port,
			        st->st_peeridentity_protocol);
#endif
		    }
		    else
		      id.id_type = st->st_myidentity_type;
		}

		MD5Update(&md5ctx, (u_char *) &id, sizeof(id));

                if (hashi)
                {
                    if (initiator)
                      MD5Update(&md5ctx, st->st_myidentity,
                                st->st_myidentity_len);
                    else
                      MD5Update(&md5ctx, st->st_peeridentity,
                                st->st_peeridentity_len);
                }
                else
                {
                    if (initiator)
                      MD5Update(&md5ctx, st->st_peeridentity,
                                st->st_peeridentity_len);
                    else
                      MD5Update(&md5ctx, st->st_myidentity,
                                st->st_myidentity_len);
                }

		MD5Final(p, &md5ctx);
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf2, HMAC_BUFSIZE);
		MD5Update(&md5ctx, p, MD5_DIGEST_SIZE);
		MD5Final(p, &md5ctx);

		break;

	    case OAKLEY_SHA:
		*length = SHA1_DIGEST_SIZE;
		p = (u_char *) calloc(SHA1_DIGEST_SIZE, sizeof(u_char));
		if (p == (u_char *) NULL)
		  exit_log("calloc() failed in get_hash()");

		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf1, HMAC_BUFSIZE);

		if (hashi)
		{
		    SHA1Update(&sha1ctx, gi, gilen);
		    SHA1Update(&sha1ctx, gr, grlen);
		    SHA1Update(&sha1ctx, st->st_icookie, COOKIE_SIZE);
		    SHA1Update(&sha1ctx, st->st_rcookie, COOKIE_SIZE);
		}
		else
		{
		    SHA1Update(&sha1ctx, gr, grlen);
		    SHA1Update(&sha1ctx, gi, gilen);
		    SHA1Update(&sha1ctx, st->st_rcookie, COOKIE_SIZE);
		    SHA1Update(&sha1ctx, st->st_icookie, COOKIE_SIZE);
		}

		/* SAp */
		SHA1Update(&sha1ctx, st->st_sa +
			   sizeof(struct isakmp_generic),
			   ntohs(isasa->isasa_length) -
			   sizeof(struct isakmp_generic));

		bzero(&id, sizeof(id));

		if (hashi)
		{
		    if (initiator)
		      id.id_type = st->st_myidentity_type;
		    else
		    {
		        id.id_type = st->st_peeridentity_type;
		        id.id_port = st->st_peeridentity_port;
		        id.id_protoid = st->st_peeridentity_protocol;
		    }
		}
		else
		{
		    if (initiator)
		    {
		        id.id_type = st->st_peeridentity_type;
		        id.id_port = st->st_peeridentity_port;
		        id.id_protoid = st->st_peeridentity_protocol;
 		    }
		    else
		      id.id_type = st->st_myidentity_type;
		}

		SHA1Update(&sha1ctx, (u_char *) &id, sizeof(id));

		if (hashi)
		{
		    if (initiator)
		      SHA1Update(&sha1ctx, st->st_myidentity,
				 st->st_myidentity_len);
		    else
		      SHA1Update(&sha1ctx, st->st_peeridentity,
				 st->st_peeridentity_len);
		}
		else
		{
		    if (initiator)
		      SHA1Update(&sha1ctx, st->st_peeridentity,
				 st->st_peeridentity_len);
		    else
		      SHA1Update(&sha1ctx, st->st_myidentity,
				 st->st_myidentity_len);
		}

		SHA1Final(p, &sha1ctx);
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf2, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, p, SHA1_DIGEST_SIZE);
		SHA1Final(p, &sha1ctx);
		break;

	    default:
		exit_log("get_hash(): unknown hash algorithm %d", st->st_hash);
	}
    }
    else
    {
	/* XXX Handle 3DES */
    }

    free(gi);
    free(gr);
    return p;
}

/*
 * Handle HDR;KE;Nr from responder. Send a HDR;IDii;HASH_I back.
 */
void
ipsecdoi_handle_i2(int sock, u_char *buffer, int length, struct sockaddr sa,
		   struct state *st)
{
    struct isakmp_hdr *isa;
    struct isakmp_generic *isag;
    MP_INT temp, temp2;
    int i, k, blocksize, padding;
    u_char *hash;
    struct identity *ide;
    des_key_schedule des_cbc_keys;

    isa = (struct isakmp_hdr *) buffer;
    if (length <= sizeof(struct isakmp_hdr) +
	2 * sizeof(struct isakmp_generic))
    {
	free(buffer);
	log(0, "too short HDR;KE;Nr packet (%d bytes) from %s, port %d",
	    length, get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    isag = (struct isakmp_generic *) (buffer + sizeof(struct isakmp_hdr));
    i = htons(isag->isag_length);
    if (i + sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic) >
	length)
    {
	free(buffer);
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    if (i - sizeof(struct isakmp_generic) < MINIMUM_PUBLIC_VALUE_SIZE)
    {
	free(buffer);
	log(0, "too short public size (%d) sent from %s, port %d",
	    i - sizeof(struct isakmp_generic), get_address(sa),
	    get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    if (isag->isag_np != ISAKMP_NEXT_NONCE)
    {
	free(buffer);
	log(0, "unexpected header %d in HDR;KE;Nr packet from %s, port %d",
	    isag->isag_np, get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    mpz_init(&temp);
    mpz_init(&temp2);

    i += sizeof(struct isakmp_hdr);
    for (k = sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic);
	 k < i; k++)
    {
	mpz_mul_ui(&temp2, &temp, 256);
	mpz_add_ui(&temp, &temp2, buffer[k]);
    }

    /* Check the Nr header */
    isag = (struct isakmp_generic *) (buffer + i);
    if (ntohs(isag->isag_length) + i != length)
    {
	free(buffer);
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    mpz_init(&(st->st_shared));
    mpz_init(&(st->st_gr));

    st->st_shared_in_use = 1;
    st->st_gr_in_use = 1;

    mpz_set(&(st->st_gr), &temp);

    mpz_powm(&(st->st_shared), &(st->st_gr), &(st->st_sec), &groupmodulo);

#ifdef DEBUG
    log(0, "public value received: ");
    mpz_out_str(stderr, 16, &(st->st_gr));
    fprintf(stderr, "\n");
    log(0, "shared secret: ");
    mpz_out_str(stderr, 16, &(st->st_shared));
    fprintf(stderr, "\n");
#endif

    /* Save Nr */
    st->st_nr = (u_char *) calloc(ntohs(isag->isag_length), sizeof(u_char));
    if (st->st_nr == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_i2()");

    bcopy(isag, st->st_nr, ntohs(isag->isag_length));

    /* Delete old retransmit event */
    delete_event(st);

    /* Free old packet */
    free(st->st_packet);

    /* Advance state */
    st->st_state = OAKLEY_MAIN_I_3;

    /* Generate SKEYID, SKEYID_A, SKEYID_D, SKEYID_E */
    generate_skeyids(st);

#ifdef DEBUG
    log(0, "Skeyid(%d): ", st->st_skeyid_len);
    for (k = 0; k < st->st_skeyid_len; k++)
      fprintf(stderr, "%02x ", st->st_skeyid[k]);
    fprintf(stderr, "\n");
    log(0, "Skeyid_d(%d): ", st->st_skeyid_d_len);
    for (k = 0; k < st->st_skeyid_d_len; k++)
      fprintf(stderr, "%02x ", st->st_skeyid_d[k]);
    fprintf(stderr, "\n");
    log(0, "Skeyid_a(%d): ", st->st_skeyid_a_len);
    for (k = 0; k < st->st_skeyid_a_len; k++)
      fprintf(stderr, "%02x ", st->st_skeyid_a[k]);
    fprintf(stderr, "\n");
    log(0, "Skeyid_e(%d): ", st->st_skeyid_e_len);
    for (k = 0; k < st->st_skeyid_e_len; k++)
      fprintf(stderr, "%02x ", st->st_skeyid_e[k]);
    fprintf(stderr, "\n");
    log(0, "IV(%d): ", st->st_iv_len);
    for (k = 0; k < st->st_iv_len; k++)
      fprintf(stderr, "%02x ", st->st_iv[k]);
    fprintf(stderr, "\n");
#endif

    /* Generate HASH_I */
    hash = get_hash(st, &i, 1, 1);

    st->st_packet_len = sizeof(struct isakmp_hdr) +
			2 * sizeof(struct isakmp_generic) + i +
			st->st_myidentity_len + sizeof(struct identity);
    /* Padding */
    switch (st->st_enc)
    {
	case OAKLEY_DES_CBC:
	case OAKLEY_3DES_CBC:
	    blocksize = DES_CBC_BLOCK_SIZE;
	    break;

	default:
	    exit_log("unknown/unsupported encryption algorithm %d specified",
		     st->st_enc);
    }

    k = (blocksize -
	 (st->st_packet_len - sizeof(struct isakmp_hdr)) % blocksize) %
	blocksize;
    st->st_packet_len += k;

    st->st_packet = (u_char *) calloc(st->st_packet_len, sizeof(u_char));
    if (st->st_packet == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_i2()");

    /* st->st_packet[st->st_packet_len - 1] = k; */
    padding = k;

#ifdef DEBUG
    log(0, "padding is %d bytes", k);
#endif

    bcopy(isa, st->st_packet, sizeof(struct isakmp_hdr));
    isa = (struct isakmp_hdr *) st->st_packet;
    isa->isa_length = htonl(st->st_packet_len);
    isa->isa_np = ISAKMP_NEXT_ID;

    /* IDii */
    isag = (struct isakmp_generic *) (st->st_packet +
				      sizeof(struct isakmp_hdr));
    isag->isag_length = htons(sizeof(struct isakmp_generic) +
			      st->st_myidentity_len + sizeof(struct identity));
    isag->isag_np = ISAKMP_NEXT_HASH;

    ide = (struct identity *) (st->st_packet + sizeof(struct isakmp_hdr) +
			       sizeof(struct isakmp_generic));

    ide->id_type = st->st_myidentity_type;

    bcopy(st->st_myidentity,
	  st->st_packet + sizeof(struct isakmp_hdr) +
	  sizeof(struct isakmp_generic) + sizeof(struct identity),
	  st->st_myidentity_len);

    /* HASH_I */
    isag = (struct isakmp_generic *) (st->st_packet +
				      sizeof(struct isakmp_hdr) +
				      ntohs(isag->isag_length));
    isag->isag_length = htons(i + sizeof(struct isakmp_generic));

    bcopy(hash, st->st_packet + st->st_packet_len - (i + padding), i);

#ifdef DEBUG
    log(0, "HASH_I sent: ");
    for (k = 0; k < i; k++)
      fprintf(stderr, "%02x ", hash[k]);
    fprintf(stderr, "\n");
#endif

    free(hash);

    isa->isa_flags |= ISAKMP_FLAG_ENCRYPTION;

    switch (st->st_enc)
    {
	case OAKLEY_DES_CBC:
	    /* XXX Do weak key detection */
	    (void) des_set_key((des_cblock *)st->st_skeyid_e, des_cbc_keys);
	    des_cbc_encrypt((des_cblock *)(st->st_packet + sizeof(struct isakmp_hdr)),
			    (des_cblock *)(st->st_packet + sizeof(struct isakmp_hdr)),
			    st->st_packet_len - sizeof(struct isakmp_hdr),
			    des_cbc_keys, (des_cblock *)st->st_iv, 1);
#ifdef DEBUG
	    log(0, "encrypting using DES CBC");
#endif

	    /* Update IV */
	    if ((st->st_iv != (u_char *) NULL) &&
		(st->st_iv_len < DES_CBC_BLOCK_SIZE))
	    {
		free(st->st_iv);
		st->st_iv = (u_char *) NULL;
	    }

	    if (st->st_iv == (u_char *) NULL)
	    {
		st->st_iv = (u_char *) calloc(DES_CBC_BLOCK_SIZE,
					      sizeof(u_char));
		if (st->st_iv == (u_char *) NULL)
		  exit_log("calloc() failed in ipsecdoi_handle_i2()");
	    }

	    st->st_iv_len = DES_CBC_BLOCK_SIZE;
	    bcopy(st->st_packet + st->st_packet_len - DES_CBC_BLOCK_SIZE,
		  st->st_iv, DES_CBC_BLOCK_SIZE);

	    break;

	    /* XXX Support more */
	default:
	    exit_log("unsupported encryption algorithm %d", st->st_enc);
    }

#ifdef DEBUG
    log(0, "new IV: ");
    for (k = 0; k < st->st_iv_len; k++)
      fprintf(stderr, "%02x ", st->st_iv[k]);
    fprintf(stderr, "\n");
#endif

    if (sendto(sock, st->st_packet, st->st_packet_len, 0, &sa, sizeof(sa)) !=
	st->st_packet_len)
      log(1, "sendto() failed in ipsecdoi_handle_i2()");
#ifdef DEBUG
    else
      log(0, "transmitted %d bytes", st->st_packet_len);
#endif

    event_schedule(EVENT_RETRANSMIT, EVENT_RETRANSMIT_DELAY, st, 0);
}

/*
 * Compute HASH(1), HASH(2) of Quick Mode.
 */
static void
compute_hash(struct state *st, u_char *packet, int length, int hash2)
{
    MD5_CTX md5ctx;
    SHA1_CTX sha1ctx;
    u_char *p, buf1[HMAC_BUFSIZE], buf2[HMAC_BUFSIZE];
    int k, ignore;
    struct isakmp_generic *isag, *isag2;

    p = packet + sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic);

    /* Ignore the ISAKMP header and the HASH(?) space */
    isag = (struct isakmp_generic *) (packet + sizeof(struct isakmp_hdr));
    ignore = ntohs(isag->isag_length) + sizeof(struct isakmp_hdr);
    isag2 = (struct isakmp_generic *) st->st_ni;

    bzero(buf1, HMAC_BUFSIZE);
    bzero(buf2, HMAC_BUFSIZE);

#ifdef DEBUG
    log(0, "compute_hash(): skipping %d bytes at begining of packet", ignore);
#endif

    /* Prepare the two pads for the HMAC */
    if (st->st_skeyid_a_len <= HMAC_BUFSIZE)
    {
	bcopy(st->st_skeyid_a, buf1, st->st_skeyid_a_len);
	bcopy(st->st_skeyid_a, buf2, st->st_skeyid_a_len);
    }
    else
    {
	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, st->st_skeyid_a, st->st_skeyid_a_len);
		MD5Final(buf1, &md5ctx);
		bcopy(buf1, buf2, MD5_DIGEST_SIZE);
		break;

	    case OAKLEY_SHA:
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, st->st_skeyid_a, st->st_skeyid_a_len);
		SHA1Final(buf1, &sha1ctx);
		bcopy(buf1, buf2, SHA1_DIGEST_SIZE);
		break;

	    default:
		exit_log("unknown/unsupported hash algorithm %d in compute_hash()", st->st_hash);
	}
    }

    for (k = 0; k < HMAC_BUFSIZE; k++)
    {
	buf1[k] ^= HMAC_IPAD;
	buf2[k] ^= HMAC_OPAD;
    }

    if (st->st_prf == 0)
    {
	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf1, HMAC_BUFSIZE);
		MD5Update(&md5ctx, (u_char *) &(st->st_msgid),
			  sizeof(st->st_msgid));
		if (hash2)
		  MD5Update(&md5ctx, st->st_ni + sizeof(struct isakmp_generic),
			    ntohs(isag2->isag_length) -
			    sizeof(struct isakmp_generic));
		MD5Update(&md5ctx, packet + ignore, length - ignore);
		MD5Final(p, &md5ctx);
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf2, HMAC_BUFSIZE);
		MD5Update(&md5ctx, p, MD5_DIGEST_SIZE);
		MD5Final(p, &md5ctx);

		break;

	    case OAKLEY_SHA:
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf1, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, (u_char *) &(st->st_msgid),
			   sizeof(st->st_msgid));
		if (hash2)
		  SHA1Update(&sha1ctx, st->st_ni +
			     sizeof(struct isakmp_generic),
			     ntohs(isag2->isag_length) -
			     sizeof(struct isakmp_generic));
		SHA1Update(&sha1ctx, packet + ignore, length - ignore);
		SHA1Final(p, &sha1ctx);
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf2, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, p, SHA1_DIGEST_SIZE);
		SHA1Final(p, &sha1ctx);

		break;

	    default:
		exit_log("unknown/unsupported hash algorithm %d specified in compute_hash()", st->st_hash);
	}
    }
    else
    {
	/* XXX Handle 3DES */
    }

#ifdef DEBUG
    log(0, "HASH(%d) computed: ", hash2 + 1);
    for (k = 0; k < ntohs(isag->isag_length) - sizeof(struct isakmp_generic);
	 k++)
      fprintf(stderr, "%02x ", p[k]);
    fprintf(stderr, "\n");
#endif
}

/*
 * Initiate quick mode.
 */
static void
initiate_quick(int sock, struct state *st, struct sockaddr_in sin2,
	       struct sockaddr_in sin3, struct sockaddr_in sin4,
  	       struct sockaddr_in sin5)
{
    struct state *nst;
    struct isakmp_hdr *isa;
    int j, k, blocksize;
    u_char *p;
    struct isakmp_generic *isag, *isag2;
    struct isakmp_sa *isasa;
    struct sockaddr_in sin;
    u_int32_t *iaddr;
    MD5_CTX md5ctx;
    SHA1_CTX sha1ctx;
    des_key_schedule des_cbc_keys;
    struct identity *id;
    nst = duplicate_state(st);

    /* NOTE: we assume 8192 is big enough to build the packet */
    nst->st_packet = (u_char *) calloc(8192, sizeof(u_char));
    if (nst->st_packet == (u_char *) NULL)
      exit_log("calloc() failed in initiate_quick()");

    nst->st_ouruser = sin2;
    nst->st_ournetmask = sin3;
    nst->st_peeruser = sin4;
    nst->st_peernetmask = sin5;

    nst->st_msgid = oakley_get_msgid(st->st_peer);
    nst->st_state = OAKLEY_QUICK_I_1;

    /* Don't move this before the oakley_get_msgid() call */
    insert_state(nst);

    /* Fix header */
    isa = (struct isakmp_hdr *) nst->st_packet;
    isa->isa_maj = ISAKMP_MAJOR_VERSION;
    isa->isa_min = ISAKMP_MINOR_VERSION;
    isa->isa_np = ISAKMP_NEXT_HASH;
    isa->isa_xchg = ISAKMP_XCHG_QUICK;
    isa->isa_length = sizeof(struct isakmp_hdr);
    isa->isa_msgid = nst->st_msgid;
    isa->isa_flags |= ISAKMP_FLAG_ENCRYPTION;
    bcopy(nst->st_icookie, isa->isa_icookie, COOKIE_SIZE);
    bcopy(nst->st_rcookie, isa->isa_rcookie, COOKIE_SIZE);

    /* HASH(1) -- first pass */
    isag = (struct isakmp_generic *) (nst->st_packet +
				      sizeof(struct isakmp_hdr));
    isag->isag_np = ISAKMP_NEXT_SA;
    if (nst->st_prf == 0)
    {
	switch (nst->st_hash)
	{
	    case OAKLEY_MD5:
		isag->isag_length = htons(MD5_DIGEST_SIZE + sizeof(struct isakmp_generic));
		break;

	    case OAKLEY_SHA:
		isag->isag_length = htons(SHA1_DIGEST_SIZE + sizeof(struct isakmp_generic));
		break;

	    default:
		exit_log("unknown hash algorithm (%d) specified in initiate_quick()", nst->st_hash);
	}
    }
    else
    {
	switch (nst->st_prf)
	{
	    case OAKLEY_PRF_3DES_CBC_MAC:
		isag->isag_length = htons(8 + sizeof(struct isakmp_generic));
		break;

	    default:
		exit_log("unknown PRF (%d) specified in initiate_quick()",
			 nst->st_prf);
	}
    }

    /* Point after HASH(1) */
    isa->isa_length += ntohs(isag->isag_length);

    bcopy(&(nst->st_peer), &sin, sizeof(sin));
    p = oakley_get_sa(&j, sin, nst->st_goal, nst);
    if (p == (u_char *) NULL)
    {
	free_state(nst);
	return;
    }

    bcopy(p, nst->st_packet + isa->isa_length, j);

    /* Fix NextPayload field */
    isasa = (struct isakmp_sa *) (nst->st_packet + isa->isa_length);
    isasa->isasa_np = ISAKMP_NEXT_NONCE;

    /* Also copy SA into state */
    nst->st_sa = p; /* in effect, alloc new, copy old, and free old */

    /* Point right after the SA */
    isa->isa_length += j;

    /* Ni */
    isag = (struct isakmp_generic *) (nst->st_packet + isa->isa_length);
    isag->isag_length = htons(sizeof(struct isakmp_generic) +
			      DEFAULTNONCESIZE);
    get_rnd_bytes(nst->st_packet + isa->isa_length +
		  sizeof(struct isakmp_generic), DEFAULTNONCESIZE);

    isa->isa_length += ntohs(isag->isag_length);

    /* Copy Ni into state */
    nst->st_ni = (u_char *) calloc(ntohs(isag->isag_length), sizeof(u_char));
    if (nst->st_ni == (u_char *) NULL)
      exit_log("calloc() failed in initiate_quick()");

    bcopy(isag, nst->st_ni, ntohs(isag->isag_length));

    /* Proxy SA */
    if (nst->st_ouruser.sin_addr.s_addr != 0)
    {
#ifdef DEBUG
	fprintf(stdout, "Sending IDui/IDur\n");
#endif
	isag->isag_np = ISAKMP_NEXT_ID;

	/* IDui */
      	isag2 = (struct isakmp_generic *) (nst->st_packet + isa->isa_length);
	isag2->isag_np = ISAKMP_NEXT_ID;
        isag2->isag_length = htons(sizeof(struct isakmp_generic) +
			 	   sizeof(struct identity) +
				   2 * sizeof(u_int32_t));

	isa->isa_length += ntohs(isag2->isag_length);
	id = (struct identity *) (((u_char *) isag2) +
				  sizeof(struct isakmp_generic));
	id->id_type = ID_IPV4_ADDR_SUBNET;
	iaddr = (u_int32_t *) ((u_char *) id + sizeof(struct identity));
	*iaddr = nst->st_ouruser.sin_addr.s_addr;
	iaddr++;
	*iaddr = nst->st_ournetmask.sin_addr.s_addr;

	/* IDur */
	isag2 = (struct isakmp_generic *) (nst->st_packet + isa->isa_length);
        isag2->isag_length = htons(sizeof(struct isakmp_generic) +
			 	   sizeof(struct identity) +
				   2 * sizeof(u_int32_t));

	isa->isa_length += ntohs(isag2->isag_length);
	id = (struct identity *) (((u_char *) isag2) +
				  sizeof(struct isakmp_generic));
	id->id_type = ID_IPV4_ADDR_SUBNET;
	iaddr = (u_int32_t *) ((u_char *) id + sizeof(struct identity));
	*iaddr = nst->st_peeruser.sin_addr.s_addr;
	iaddr++;
	*iaddr = nst->st_peernetmask.sin_addr.s_addr;
    }

    /* Compute the Phase 2 IV */

    if (nst->st_prf == 0)
    {
	switch (nst->st_hash)
	{
	    case OAKLEY_MD5:
		nst->st_iv_len = MD5_DIGEST_SIZE;
		nst->st_iv = (u_char *) calloc(nst->st_iv_len,
					       sizeof(u_char));
		if (nst->st_iv == (u_char *) NULL)
		  exit_log("calloc() failed in initiate_quick()");

		MD5Init(&md5ctx);
		MD5Update(&md5ctx, st->st_lastblock, st->st_lastblock_len);
		MD5Update(&md5ctx, (u_char *) &(nst->st_msgid),
			  sizeof(nst->st_msgid));
		MD5Final(nst->st_iv, &md5ctx);

		break;

	    case OAKLEY_SHA:
		nst->st_iv_len = SHA1_DIGEST_SIZE;
		nst->st_iv = (u_char *) calloc(nst->st_iv_len,
					       sizeof(u_char));
		if (nst->st_iv == (u_char *) NULL)
		  exit_log("calloc() failed in initiate_quick()");

		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, st->st_lastblock, st->st_lastblock_len);
		SHA1Update(&sha1ctx, (u_char *) &(nst->st_msgid),
			   sizeof(nst->st_msgid));
		SHA1Final(nst->st_iv, &sha1ctx);

		break;

	    default:
		exit_log("unknown/unsupported hash algorithm %d specified in initiate_quick()", nst->st_hash);
	}
    }
    else
    {
	/* XXX 3DES-CBC MAC */
    }

#ifdef DEBUG
    log(0, "computed Phase 2 IV: ");
    for (k = 0; k < nst->st_iv_len; k++)
      fprintf(stderr, "%02x ", nst->st_iv[k]);
    fprintf(stderr, "\n");
#endif

    /* HASH(1) -- compute */

    compute_hash(nst, nst->st_packet, isa->isa_length, 0);

    /* Padding */
    switch (nst->st_enc)
    {
	case OAKLEY_DES_CBC:
	case OAKLEY_3DES_CBC:
	    blocksize = DES_CBC_BLOCK_SIZE;
	    break;

	default:
	    exit_log("unknown/unsupported encryption algorithm %d specified",
		     nst->st_enc);
    }

    nst->st_packet_len = isa->isa_length;

#ifdef DEBUG
    log(0, "Packet length %d", nst->st_packet_len);
#endif

    k = (blocksize -
	 (nst->st_packet_len - sizeof(struct isakmp_hdr)) % blocksize) %
	blocksize;

#ifdef DEBUG
    log(0, "padding is %d bytes", k);
#endif

    isa->isa_length += k;		/* Add padding */

    /* Update lengths */
    nst->st_packet_len = isa->isa_length;
    isa->isa_length = htonl(isa->isa_length);

    /* Encrypt */
    switch (nst->st_enc)
    {
	case OAKLEY_DES_CBC:
	    /* XXX Do weak key detection */
	    (void) des_set_key((des_cblock *)nst->st_skeyid_e, des_cbc_keys);
	    des_cbc_encrypt((des_cblock *)(nst->st_packet + sizeof(struct isakmp_hdr)),
			    (des_cblock *)(nst->st_packet + sizeof(struct isakmp_hdr)),
			    nst->st_packet_len - sizeof(struct isakmp_hdr),
			    des_cbc_keys, (des_cblock *)nst->st_iv, 1);


#ifdef DEBUG
	    log(0, "encrypting using DES CBC");
#endif

	    /* Keep last block as IV */
	    if ((nst->st_iv != (u_char *) NULL) &&
		(nst->st_iv_len < DES_CBC_BLOCK_SIZE))
	    {
		free(nst->st_iv);
		nst->st_iv = (u_char *) NULL;
	    }

	    if (nst->st_iv == (u_char *) NULL)
	    {
		nst->st_iv = (u_char *) calloc(DES_CBC_BLOCK_SIZE,
					       sizeof(u_char));
		if (nst->st_iv == (u_char *) NULL)
		  exit_log("calloc() failed in initiate_quick()");
	    }

	    bcopy(nst->st_packet + nst->st_packet_len - DES_CBC_BLOCK_SIZE ,
		  nst->st_iv, DES_CBC_BLOCK_SIZE);
	    nst->st_iv_len = DES_CBC_BLOCK_SIZE;

	    break;

	    /* XXX Support more */
	default:
	    exit_log("unsupported encryption algorithm %d", nst->st_enc);
    }

#ifdef DEBUG
    log(0, "new IV: ");
    for (k = 0; k < nst->st_iv_len; k++)
      fprintf(stderr, "%02x ", nst->st_iv[k]);
    fprintf(stderr, "\n");
#endif

    if (sendto(sock, nst->st_packet, nst->st_packet_len, 0, &(nst->st_peer),
	       sizeof(nst->st_peer)) != nst->st_packet_len)
      log(1, "sendto() failed in initiate_quick()");
#ifdef DEBUG
    else
      log(0, "transmitted %d bytes", nst->st_packet_len);
#endif

    event_schedule(EVENT_RETRANSMIT, EVENT_RETRANSMIT_DELAY, nst, 0);
}

/*
 * Handle HDR;IDir;HASH_R from responder.
 */
void
ipsecdoi_handle_i3(int sock, u_char *buffer, int length, struct sockaddr sa,
		   struct state *st)
{
    struct isakmp_hdr *isa;
    struct isakmp_generic *isag;
    int i, k;
    struct identity *id;
    u_char *hash;
    struct sockaddr_in sin;

    isa = (struct isakmp_hdr *) buffer;
    if (length <= sizeof(struct isakmp_hdr) +
	2 * sizeof(struct isakmp_generic) + sizeof(struct identity))
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    if (isa->isa_np != ISAKMP_NEXT_ID)
    {
	log(0, "bad packet received from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    i = sizeof(struct isakmp_hdr);

    /* IDir */
    isag = (struct isakmp_generic *) (buffer + i);
    if (ntohs(isag->isag_length) + i >= length)
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    i += sizeof(struct isakmp_generic);
    if (isag->isag_np != ISAKMP_NEXT_HASH)
    {
	log(0, "bad packet received from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    id = (struct identity *) (buffer + i);

#ifdef 0
    /* XXX Check for valid ID types ? */
    if ((id->id_port != 0) || (id->id_protoid))
    {
	log(0, "port or protocol id in ID not zero (%d/%d)", id->id_port,
	    id->id_protoid);
	free(buffer);
	/* XXX Could send notification back */
	return;
    }
#endif

    i += sizeof(struct identity);
    st->st_peeridentity_protocol = id->id_protoid;
    st->st_peeridentity_port = id->id_port;
    st->st_peeridentity_type = id->id_type;
    st->st_peeridentity_len = ntohs(isag->isag_length) -
			      (sizeof(struct isakmp_generic) +
			       sizeof(struct identity));
    st->st_peeridentity = (u_char *) calloc(st->st_peeridentity_len,
					    sizeof(u_char));
    if (st->st_peeridentity == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_i3()");

    bcopy(buffer + i, st->st_peeridentity, st->st_peeridentity_len);
    i += st->st_peeridentity_len;

#ifdef DEBUG
    log(0, "IDir type is %d, length %d", id->id_type,
	st->st_peeridentity_len);
    if (id->id_type == ID_IPV4_ADDR)
    {
	bcopy(st->st_peeridentity, &(sin.sin_addr), sizeof(sin));
	log(0, "IDir is %s", inet_ntoa(sin.sin_addr));
    }
#endif

    /* HASH_R */
    isag = (struct isakmp_generic *) (buffer + i);

    if ((isag->isag_np != ISAKMP_NEXT_NONE) ||
	(ntohs(isag->isag_length) + i != length))
    {
	log(0, "bad packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	free(st->st_peeridentity);	/* Rollback */
	/* XXX Could send notification back */
	return;
    }

    i += sizeof(struct isakmp_generic);

    /* Compute HASH_R */
    hash = get_hash(st, &k, 0, 1);

    if (length - i != k)
    {
	log(0, "computed and received HASH_R lengths don't match (%d/%d)",
	    length - i, k);
	free(buffer);
	free(hash);
	free(st->st_peeridentity);
	/* XXX Could send notification back */
	return;
    }

#ifdef DEBUG
    log(0, "computed HASH_R: ");
    for (k = 0; k < length - i; k++)
      fprintf(stderr, "%02x ", hash[k]);
    fprintf(stderr, "\n");
    log(0, "received HASH_R: ");
    for (k = 0; k < length - i; k++)
      fprintf(stderr, "%02x ", buffer[i + k]);
    fprintf(stderr, "\n");
#endif

    if (bcmp(hash, buffer + i, length - i))
    {
	log(0, "received and computed HASH_R values don't match");
	free(buffer);
	free(hash);
	free(st->st_peeridentity);
	/* XXX Could send notification back */
	return;
    }
#ifdef DEBUG
    else
      log (0, "HASH_R verified");
#endif

    /* Free old packet */
    free(st->st_packet);

    /* Free computed HASH_I */
    free(hash);

    /* Free received packet */
    free(buffer);

    /* Delete pending event */
    delete_event(st);

    /* Advance state */
    st->st_state = OAKLEY_MAIN_I_4;

    if (st->st_goal != 0)
    {
#ifdef DEBUG
	log(0, "Doing Quick Mode with %s, port %d, goal %d",
	    get_address(sa), get_port(sa), st->st_goal);
#endif
	initiate_quick(sock, st, st->st_ouruser, st->st_ournetmask,
		       st->st_peeruser, st->st_peernetmask);
    }

    /* Schedule for SA expiration */
    if (st->st_expire != 0)
      event_schedule(EVENT_SA_EXPIRE, st->st_expire, st, 0);
    else
      event_schedule(EVENT_SA_EXPIRE, DEFAULT_SA_EXPIRE, st, 0);
}

/*
 * Handle HDR;KE;Ni from Initiator. Send a HDR;KE;Nr back.
 */
void
ipsecdoi_handle_r1(int sock, u_char *buffer, int length, struct sockaddr sa,
		   struct state *st)
{
    struct isakmp_hdr *isa;
    struct isakmp_generic *isag;
    u_char *packet;
    int i, k;
    u_char tmp[LOCALSECRETSIZE];
    MP_INT temp, temp2;

    isa = (struct isakmp_hdr *) buffer;

    if (length <= sizeof(struct isakmp_hdr) +
	2 * sizeof(struct isakmp_generic))
    {
	free(buffer);
	log(0, "too short HDR;KE;Ni packet (%d bytes) from %s, port %d",
	    length, get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    isag = (struct isakmp_generic *) (buffer + sizeof(struct isakmp_hdr));
    i = htons(isag->isag_length);
    if (i + sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic) >
	length)
    {
	free(buffer);
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    if (i - sizeof(struct isakmp_generic) < MINIMUM_PUBLIC_VALUE_SIZE)
    {
	free(buffer);
	log(0, "too short public size (%d) sent from %s, port %d",
	    i - sizeof(struct isakmp_generic), get_address(sa),
	    get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    if (isag->isag_np != ISAKMP_NEXT_NONCE)
    {
	free(buffer);
	log(0, "unexpected header %d in HDR;KE;Ni packet from %s, port %d",
	    isag->isag_np, get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    mpz_init(&temp);
    mpz_init(&temp2);

    i += sizeof(struct isakmp_hdr);
    for (k = sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic);
	 k < i; k++)
    {
	mpz_mul_ui(&temp2, &temp, 256);
	mpz_add_ui(&temp, &temp2, buffer[k]);
    }

    /* Check the Ni header */
    isag = (struct isakmp_generic *) (buffer + i);
    if (ntohs(isag->isag_length) + i != length)
    {
	free(buffer);
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    mpz_init(&(st->st_gi));
    mpz_init(&(st->st_gr));
    mpz_init(&(st->st_sec));
    mpz_init(&(st->st_shared));

    st->st_gi_in_use = 1;
    st->st_gr_in_use = 1;
    st->st_sec_in_use = 1;
    st->st_shared_in_use = 1;

    mpz_set(&(st->st_gi), &temp);

#ifdef DEBUG
    log(0, "public value received:");
    mpz_out_str(stderr, 16, &temp);
    fprintf(stderr, "\n");
#endif

    get_rnd_bytes(tmp, LOCALSECRETSIZE);

    mpz_set_ui(&temp, 0);

    for (k = 0; k < LOCALSECRETSIZE; k++)
    {
	mpz_mul_ui(&temp2, &temp, 256);
	mpz_add_ui(&temp, &temp2, tmp[k]);
    }

    mpz_set(&(st->st_sec), &temp);

    mpz_powm(&(st->st_gr), &groupgenerator, &(st->st_sec), &groupmodulo);

    mpz_powm(&(st->st_shared), &(st->st_gi), &(st->st_sec), &groupmodulo);

#ifdef DEBUG
    log(0, "our secret value: ");
    mpz_out_str(stderr, 16, &(st->st_sec));
    fprintf(stderr, "\n");
    log(0, "our public value: ");
    mpz_out_str(stderr, 16, &(st->st_gr));
    fprintf(stderr, "\n");
    log(0, "shared secret: ");
    mpz_out_str(stderr, 16, &(st->st_shared));
    fprintf(stderr, "\n");
#endif

    /* Save Ni */
    st->st_ni = (u_char *) calloc(ntohs(isag->isag_length), sizeof(u_char));
    if (st->st_ni == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_r1()");

    bcopy(isag, st->st_ni, ntohs(isag->isag_length));

    /* Create Nr */
    st->st_nr = (u_char *) calloc(DEFAULTNONCESIZE +
				  sizeof(struct isakmp_generic),
				  sizeof(u_char));
    if (st->st_nr == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_r1()");

    isag = (struct isakmp_generic *) st->st_nr;
    isag->isag_length = ntohs(DEFAULTNONCESIZE +
			      sizeof(struct isakmp_generic));

    get_rnd_bytes(st->st_nr + sizeof(struct isakmp_generic), DEFAULTNONCESIZE);

    i = sizeof(struct isakmp_hdr) + 2 * sizeof(struct isakmp_generic) +
	(mpz_sizeinbase(&(st->st_gr), 16) + 1) / 2 + DEFAULTNONCESIZE;

    packet = (u_char *) calloc(i, sizeof(u_char));
    if (packet == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_r1()");

    bcopy(isa, packet, sizeof(struct isakmp_hdr));

    isa = (struct isakmp_hdr *) packet;
    isag = (struct isakmp_generic *) (packet + sizeof(struct isakmp_hdr));
    isag->isag_np = ISAKMP_NEXT_NONCE;
    isa->isa_length = htonl(i);
    isag->isag_length = htons((mpz_sizeinbase(&(st->st_gr), 16) + 1) / 2 +
			      sizeof(struct isakmp_generic));

    mpz_set(&temp, &(st->st_gr));

    i -= (sizeof(struct isakmp_generic) + DEFAULTNONCESIZE);

    for (k = i - 1;
	 k >= sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic);
	 k--)
    {
	packet[k] = mpz_mdivmod_ui(&temp2, NULL, &temp, 256);
	mpz_set(&temp, &temp2);
    }

    mpz_clear(&temp);
    mpz_clear(&temp2);

    isag = (struct isakmp_generic *) st->st_nr;
    bcopy(st->st_nr, packet + i, ntohs(isag->isag_length));

    /* Advance state */
    st->st_state = OAKLEY_MAIN_R_2;

    /* Delete previous CLEANUP event */
    delete_event(st);

    /* Free received packet */
    free(buffer);

    /* No need to free old packet, since we didn't keep it */

    st->st_packet = packet;
    st->st_packet_len = i + DEFAULTNONCESIZE + sizeof(struct isakmp_generic);

    if (sendto(sock, packet, st->st_packet_len, 0, &sa, sizeof(sa)) !=
	st->st_packet_len)
      log(1, "sendto() failed in ipsecdoi_handle_r1()");
#ifdef DEBUG
    else
      log(0, "transmitted %d bytes", st->st_packet_len);
#endif

    generate_skeyids(st);

#ifdef DEBUG
    log(0, "Skeyid(%d): ", st->st_skeyid_len);
    for (k = 0; k < st->st_skeyid_len; k++)
      fprintf(stderr, "%02x ", st->st_skeyid[k]);
    fprintf(stderr, "\n");
    log(0, "Skeyid_d(%d): ", st->st_skeyid_d_len);
    for (k = 0; k < st->st_skeyid_d_len; k++)
      fprintf(stderr, "%02x ", st->st_skeyid_d[k]);
    fprintf(stderr, "\n");
    log(0, "Skeyid_a(%d): ", st->st_skeyid_a_len);
    for (k = 0; k < st->st_skeyid_a_len; k++)
      fprintf(stderr, "%02x ", st->st_skeyid_a[k]);
    fprintf(stderr, "\n");
    log(0, "Skeyid_e(%d): ", st->st_skeyid_e_len);
    for (k = 0; k < st->st_skeyid_e_len; k++)
      fprintf(stderr, "%02x ", st->st_skeyid_e[k]);
    fprintf(stderr, "\n");
    log(0, "IV(%d): ", st->st_iv_len);
    for (k = 0; k < st->st_iv_len; k++)
      fprintf(stderr, "%02x ", st->st_iv[k]);
    fprintf(stderr, "\n");
#endif

    event_schedule(EVENT_CLEANUP, EVENT_CLEANUP_DELAY, st, 0);
}

/*
 * Handle HDR;IDii;HASH_I from initiator. Send a HDR;IDir;HASH_R back.
 */
void
ipsecdoi_handle_r2(int sock, u_char *buffer, int length, struct sockaddr sa,
		   struct state *st)
{
    struct isakmp_hdr *isa;
    struct isakmp_generic *isag;
    struct identity *id;
    u_char *hash;
    int i, k, blocksize, padding;
    struct sockaddr_in sin;
    des_key_schedule des_cbc_keys;

    isa = (struct isakmp_hdr *) buffer;
    if (length <= sizeof(struct isakmp_hdr) +
	2 * sizeof(struct isakmp_generic) + sizeof(struct identity))
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    if (isa->isa_np != ISAKMP_NEXT_ID)
    {
	log(0, "bad packet received from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    i = sizeof(struct isakmp_hdr);

    /* IDii */
    isag = (struct isakmp_generic *) (buffer + i);
    if (ntohs(isag->isag_length) + i >= length)
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    i += sizeof(struct isakmp_generic);
    if (isag->isag_np != ISAKMP_NEXT_HASH)
    {
	log(0, "bad packet received from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    id = (struct identity *) (buffer + i);

    /* XXX Check for valid ID types ? */

    if ((id->id_port != 0) || (id->id_protoid))
    {
	log(0, "port or protocol id in ID not zero (%d/%d)", id->id_port,
	    id->id_protoid);
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    i += sizeof(struct identity);
    st->st_peeridentity_type = id->id_type;
    st->st_peeridentity_len = ntohs(isag->isag_length) -
			      (sizeof(struct isakmp_generic) +
			       sizeof(struct identity));
    st->st_peeridentity = (u_char *) calloc(st->st_peeridentity_len,
					    sizeof(u_char));
    if (st->st_peeridentity == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_r2()");

    bcopy(buffer + i, st->st_peeridentity, st->st_peeridentity_len);
    i += st->st_peeridentity_len;

#ifdef DEBUG
    log(0, "IDii type is %d, length %d", id->id_type,
	st->st_peeridentity_len);
    if (id->id_type == ID_IPV4_ADDR)
    {
	bcopy(st->st_peeridentity, &(sin.sin_addr), sizeof(sin));
	log(0, "IDii is %s", inet_ntoa(sin.sin_addr));
    }
#endif

    /* HASH_I */
    isag = (struct isakmp_generic *) (buffer + i);

    if ((isag->isag_np != ISAKMP_NEXT_NONE) ||
	(ntohs(isag->isag_length) + i != length))
    {
	log(0, "bad packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	free(st->st_peeridentity);	/* Rollback */
	/* XXX Could send notification back */
	return;
    }

    i += sizeof(struct isakmp_generic);

    /* Compute HASH_I */
    hash = get_hash(st, &k, 1, 0);

    if (length - i != k)
    {
	log(0, "computed and received HASH_I lengths don't match (%d/%d)",
	    length - i, k);
	free(buffer);
	free(hash);
	free(st->st_peeridentity);
	/* XXX Could send notification back */
	return;
    }

#ifdef DEBUG
    log(0, "computed HASH_I: ");
    for (k = 0; k < length - i; k++)
      fprintf(stderr, "%02x ", hash[k]);
    fprintf(stderr, "\n");
    log(0, "received HASH_I: ");
    for (k = 0; k < length - i; k++)
      fprintf(stderr, "%02x ", buffer[i + k]);
    fprintf(stderr, "\n");
#endif

    if (bcmp(hash, buffer + i, length - i))
    {
	log(0, "received and computed HASH_I values don't match");
	free(buffer);
	free(hash);
	free(st->st_peeridentity);
	/* XXX Could send notification back */
	return;
    }
#ifdef DEBUG
    else
      log (0, "HASH_I verified");
#endif

    /* Free old packet */
    free(st->st_packet);

    /* Free computed HASH_I */
    free(hash);

    st->st_packet_len = sizeof(struct isakmp_hdr) +
			2 * sizeof(struct isakmp_generic) + k +
			st->st_myidentity_len + sizeof(struct identity);

    /* Padding */
    switch (st->st_enc)
    {
	case OAKLEY_DES_CBC:
	case OAKLEY_3DES_CBC:
	    blocksize = DES_CBC_BLOCK_SIZE;
	    break;

	default:
	    exit_log("unknown/unsupported encryption algorithm %d specified",
		     st->st_enc);
    }

    k = (blocksize -
	 (st->st_packet_len - sizeof(struct isakmp_hdr)) % blocksize) %
	blocksize;

    st->st_packet_len += k;

#ifdef DEBUG
    log(0, "padding is %d bytes", k);
#endif

    st->st_packet = (u_char *) calloc(st->st_packet_len, sizeof(u_char));
    if (st->st_packet == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_r2()");

    /* st->st_packet[st->st_packet_len - 1] = k; */
    padding = k;

    bcopy(buffer, st->st_packet, sizeof(struct isakmp_hdr));
    isa = (struct isakmp_hdr *) st->st_packet;
    isa->isa_length = htonl(st->st_packet_len);

    /* IDir */
    i = sizeof(struct isakmp_hdr);
    isag = (struct isakmp_generic *) (st->st_packet + i);
    i += sizeof(struct isakmp_generic);
    id = (struct identity *) (st->st_packet + i);
    i += sizeof(struct identity);

    isag->isag_np = ISAKMP_NEXT_HASH;
    isag->isag_length = htons(sizeof(struct isakmp_generic) +
			      sizeof(struct identity) +
			      st->st_myidentity_len);

    id->id_type = st->st_myidentity_type;
    bcopy(st->st_myidentity, st->st_packet + i, st->st_myidentity_len);
    i += st->st_myidentity_len;

    /* Get HASH_R */
    hash = get_hash(st, &k, 0, 0);

    /* HASH_R */
    isag = (struct isakmp_generic *) (st->st_packet + i);
    i += sizeof(struct isakmp_generic);
    isag->isag_length = htons(sizeof(struct isakmp_generic) + k);
    bcopy(hash, st->st_packet + i, k);

#ifdef DEBUG
    log(0, "sending HASH_R: ");
    for (i = 0; i < k; i++)
      fprintf(stderr, "%02x ", hash[i]);
    fprintf(stderr, "\n");
#endif

    /* Free computed HASH_R */
    free(hash);

    /* Advance state */
    st->st_state = OAKLEY_MAIN_R_3;

    switch (st->st_enc)
    {
	case OAKLEY_DES_CBC:
	    /* XXX Do weak key detection */
	    (void) des_set_key((des_cblock *)st->st_skeyid_e, des_cbc_keys);
	    des_cbc_encrypt((des_cblock *)(st->st_packet + sizeof(struct isakmp_hdr)),
			    (des_cblock *)(st->st_packet + sizeof(struct isakmp_hdr)),
			    st->st_packet_len - sizeof(struct isakmp_hdr),
			    des_cbc_keys, (des_cblock *)st->st_iv, 1);


#ifdef DEBUG
	    log(0, "encrypting using DES CBC");
#endif

	    /* Last block of Phase 1, kept for Phase 2 IV generation */
	    st->st_lastblock = (u_char *) calloc(DES_CBC_BLOCK_SIZE,
						 sizeof(u_char));
	    if (st->st_lastblock == (u_char *) NULL)
	      exit_log("calloc() failed in ipsecdoi_handle_r2()");


	    bcopy(st->st_packet + st->st_packet_len - DES_CBC_BLOCK_SIZE,
		  st->st_lastblock, DES_CBC_BLOCK_SIZE);
	    st->st_lastblock_len = DES_CBC_BLOCK_SIZE;

	    /* Keep last block as IV */
	    if ((st->st_iv != (u_char *) NULL) &&
		(st->st_iv_len < DES_CBC_BLOCK_SIZE))
	    {
		free(st->st_iv);
		st->st_iv = (u_char *) NULL;
	    }

	    if (st->st_iv == (u_char *) NULL)
	    {
		st->st_iv = (u_char *) calloc(DES_CBC_BLOCK_SIZE,
					      sizeof(u_char));
		if (st->st_iv == (u_char *) NULL)
		  exit_log("calloc() failed in ipsecdoi_handle_r2()");
	    }

	    bcopy(st->st_packet + st->st_packet_len - DES_CBC_BLOCK_SIZE ,
		  st->st_iv, DES_CBC_BLOCK_SIZE);
	    st->st_iv_len = DES_CBC_BLOCK_SIZE;

	    break;

	    /* XXX Support more */
	default:
	    exit_log("unsupported encryption algorithm %d", st->st_enc);
    }

#ifdef DEBUG
    log(0, "last encrypted block of Phase 1: ");
    for (k = 0; k < st->st_lastblock_len; k++)
      fprintf(stderr, "%02x ", st->st_lastblock[k]);
    fprintf(stderr, "\n");
    log(0, "new IV: ");
    for (k = 0; k < st->st_iv_len; k++)
      fprintf(stderr, "%02x ", st->st_iv[k]);
    fprintf(stderr, "\n");
#endif

    /* Get rid of received packet */
    free(buffer);

    /* Delete pending CLEANUP event */
    delete_event(st);

    if (sendto(sock, st->st_packet, st->st_packet_len, 0, &sa, sizeof(sa))
	!= st->st_packet_len)
      log(1, "sendto() failed in ipsecdoi_handle_r2()");
#ifdef DEBUG
    else
      log(0, "transmitted %d bytes", st->st_packet_len);
#endif

    /* Expiration of SA */
    if (st->st_expire != 0)
      event_schedule(EVENT_SA_EXPIRE, st->st_expire, st, 0);
    else
      event_schedule(EVENT_SA_EXPIRE, DEFAULT_SA_EXPIRE, st, 0);
}

/*
 * Handle first message of Phase 2 -- Quick Mode.
 */
void
ipsecdoi_handle_quick_r1(int sock, u_char *buffer, int length,
			 struct sockaddr sa, struct state *st)
{
    struct isakmp_hdr *isa;
    struct isakmp_generic *isag, *isag2;
    struct isakmp_sa *isasa, *isas;
    struct sockaddr_in sin;
    int i, k, blocksize, j, idlength;
    u_char *hashval, *p, *idstore;
    des_key_schedule des_cbc_keys;
    u_int32_t *iaddr;

    struct identity *id;

    isa = (struct isakmp_hdr *) buffer;

    if (isa->isa_np != ISAKMP_NEXT_HASH)
    {
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    isag = (struct isakmp_generic *) (buffer + sizeof(struct isakmp_hdr));

    if (isag->isag_np != ISAKMP_NEXT_SA)
    {
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    if (length <= sizeof(struct isakmp_hdr) +
	2 * sizeof(struct isakmp_generic) + sizeof(struct isakmp_sa))
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    i = sizeof(struct isakmp_hdr);

    /* Check HASH(1) */
    if (ntohs(isag->isag_length) + sizeof(struct isakmp_hdr) >=
	length)
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    i += sizeof(struct isakmp_generic);

    /* Copy the hash */
    k = ntohs(isag->isag_length) - sizeof(struct isakmp_generic);
    hashval = (u_char *) calloc(k, sizeof(u_char));
    if (hashval == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_r1()");

    bcopy(buffer + i, hashval, k);

#ifdef DEBUG
    log(0, "received HASH(1): ");
    for (j = 0; j < k; j++)
      fprintf(stderr, "%02x ", buffer[i + j]);
    fprintf(stderr, "\n");
#endif

    compute_hash(st, buffer, length, 0);

    if (bcmp(hashval, buffer + i, k))
    {
	free(buffer);
	free(hashval);
	log(0, "HASH(1) from %s, port %d does not verify", get_address(sa),
	    get_port(sa));
	/* XXX Could send notification back */
	return;
    }

#ifdef DEBUG
    log(0, "HASH(1) verified");
#endif

    /* Point at the SA payload */
    i += k;

    /* Authentication succeeded */
    free(hashval);

    isasa = (struct isakmp_sa *) (buffer + i);
    if (i + ntohs(isasa->isasa_length) >= length)
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    if (isasa->isasa_np != ISAKMP_NEXT_NONCE)
    {
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    if (ipsecdoi_ipsec_sa_copy(buffer, length, sa, st) == 0)
    {
	free(buffer);
	free_state(st);
	log(0, "no acceptable proposal from %s, port %d",
	    get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    /* Ni payload */
    isag = (struct isakmp_generic *) (buffer + i + ntohs(isasa->isasa_length));

    if (ntohs(isag->isag_length) + i + ntohs(isasa->isasa_length) > length)
    {
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    isag2 = isag;	/* Keep this around */

    /* Handle additional payloads...right now we error */
    switch (isag->isag_np)
    {
	case ISAKMP_NEXT_NONE:
	    break;

	case ISAKMP_NEXT_ID:
	    idlength = length - (i + ntohs(isasa->isasa_length) +
		         	 ntohs(isag->isag_length));
#ifdef DEBUG
	    log(0, "keeping %d bytes of user identities", idlength);
#endif
	    isag = (struct isakmp_generic *) (buffer + i +
					      ntohs(isasa->isasa_length) +
					      ntohs(isag->isag_length));
	    idstore = (u_char *) isag;
	    if (isag->isag_np != ISAKMP_NEXT_ID)
	    {
		log(0, "malformed packet from %s, port %d", get_address(sa),
	    	    get_port(sa));
		free(buffer);
		/* XXX Could send notification back */
		return;
	    }

	    /* IDui */
	    id = (struct identity *) (isag + 1);

	    /* XXX support more */
	    switch (id->id_type)
	    {
		case ID_IPV4_ADDR:
		    iaddr = (u_int32_t *) (id + 1);
		    st->st_peeruser.sin_addr.s_addr = *iaddr;
		    st->st_peernetmask.sin_addr.s_addr = 0xffffffff;
#ifdef DEBUG
		    log(0, "peer user is IP address %s",
			inet_ntoa(st->st_peeruser.sin_addr));
#endif
		    break;

		case ID_IPV4_ADDR_SUBNET:
		    iaddr = (u_int32_t *) (id + 1);
		    st->st_peeruser.sin_addr.s_addr = *iaddr;
		    iaddr++;
		    st->st_peernetmask.sin_addr.s_addr = *iaddr;
#ifdef DEBUG
		    log(0, "peer user is IP subnet with address %s...",
			inet_ntoa(st->st_peeruser.sin_addr));
		    log(0, "...and netmask %s",
			inet_ntoa(st->st_peernetmask.sin_addr));
#endif
		    break;

		default:
		    log(0, "bad id type %d from %s, port %d", id->id_type,
			get_address(sa), get_port(sa));
		    free(buffer);
		    /* XXX Could send notification back */
		    return;
	    }

	    /* IDur */
	    isag = (struct isakmp_generic *) ((u_char *) isag +
					      ntohs(isag->isag_length));
	    if (isag->isag_np != ISAKMP_NEXT_NONE)
	    {
                log(0, "malformed packet from %s, port %d", get_address(sa),
                    get_port(sa));
                free(buffer);
                /* XXX Could send notification back */
                return;
            }

            id = (struct identity *) (isag + 1);
	    st->st_ouruserport = id->id_port;
	    st->st_ouruserprotoid = id->id_protoid;

            /* XXX support more */
            switch (id->id_type)
            {
                case ID_IPV4_ADDR:
                    iaddr = (u_int32_t *) (id + 1);
                    st->st_ouruser.sin_addr.s_addr = *iaddr;
                    st->st_ournetmask.sin_addr.s_addr = 0xffffffff;
#ifdef DEBUG
                    log(0, "our user is IP address %s",
                        inet_ntoa(st->st_ouruser.sin_addr));
#endif
                    break;

                case ID_IPV4_ADDR_SUBNET:
                    iaddr = (u_int32_t *) (id + 1);
                    st->st_ouruser.sin_addr.s_addr = *iaddr;
                    iaddr++;
                    st->st_ournetmask.sin_addr.s_addr = *iaddr;
#ifdef DEBUG
                    log(0, "our user is IP subnet with address %s...",
                        inet_ntoa(st->st_ouruser.sin_addr));
                    log(0, "...and netmask %s",
                        inet_ntoa(st->st_ournetmask.sin_addr));
#endif
                    break;

                default:
                    log(0, "bad id type %d from %s, port %d", id->id_type,
                        get_address(sa), get_port(sa));
                    free(buffer);
                    /* XXX Could send notification back */
                    return;
            }

	    isag = (struct isakmp_generic *) idstore;
	    idstore = (u_char *) calloc(idlength, sizeof(u_char));
	    if (idstore == (u_char *) NULL)
      	      exit_log("calloc() failed in ipsecdoi_handle_quick_r1()");

            bcopy(isag, idstore, idlength);
	    break;

	case ISAKMP_NEXT_KE:
	    /* XXX unsupported now -- fall through */
	default:
	    log(0,
		"unsupported payload received in Quick Mode from %s, port %d",
	     	get_address(sa), get_port(sa));
	    free(buffer);
	    /* XXX Could send notification back */
	    return;
    }

    /* Copy Ni into the state */
    st->st_ni = (u_char *) calloc(ntohs(isag2->isag_length), sizeof(u_char));
    if (st->st_ni == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_r1()");
    bcopy(isag2, st->st_ni, ntohs(isag2->isag_length));

    /* The reply packet can't be larger than the original request */
    st->st_packet = (u_char *) calloc(length, sizeof(u_char));
    if (st->st_packet == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_r1()");

    /* Copy header */
    bcopy(buffer, st->st_packet, sizeof(struct isakmp_hdr));
    isa = (struct isakmp_hdr *) st->st_packet;
    isa->isa_length = 0;
    i = sizeof(struct isakmp_hdr);

    /* HASH(2) -- first pass */
    isag = (struct isakmp_generic *) (st->st_packet + i);
    isag->isag_np = ISAKMP_NEXT_SA;

    if (st->st_prf == 0)
    {
	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		isag->isag_length = htons(MD5_DIGEST_SIZE + sizeof(struct isakmp_generic));
		break;

	    case OAKLEY_SHA:
		isag->isag_length = htons(SHA1_DIGEST_SIZE + sizeof(struct isakmp_generic));
		break;

	    default:
		exit_log("unknown hash algorithm (%d) specified in ipsecdoi_handle_quick_r1()", st->st_hash);
	}
    }
    else
    {
	switch (st->st_prf)
	{
	    case OAKLEY_PRF_3DES_CBC_MAC:
		isag->isag_length = htons(8 + sizeof(struct isakmp_generic));
		break;

	    default:
		exit_log("unknown PRF (%d) specified in ipsecdoi_handle_quick_r1()",
			 st->st_prf);
	}
    }

    i += ntohs(isag->isag_length);

    /* SA handling */
    isas = (struct isakmp_sa *) (st->st_packet + i);
    isas->isasa_np = ISAKMP_NEXT_NONCE;
    isas->isasa_doi = ntohl(ISAKMP_DOI_IPSEC);
    k = sizeof(struct isakmp_hdr) + sizeof(struct isakmp_sa) +
	ntohs(isag->isag_length) + IPSEC_DOI_SITUATION_LENGTH;
    st->st_packet[k - 1] = SIT_IDENTITY_ONLY;

    /* Set up a reverse path SA with similar characteristics */
    switch (st->st_peer_protoid)
    {
	case PROTO_IPSEC_AH:
	    st->st_goal = GOAL_AUTHENTICATE;
	    break;

	case PROTO_IPSEC_ESP:
	    st->st_goal = GOAL_ENCRYPT;
	    break;

	default:
	    exit_log("unknown protocol %d negotiated", st->st_peer_protoid);
    }

    /* Just do what the peer does */
    if (st->st_peer_encapsulation_mode == ENCAPSULATION_MODE_TUNNEL)
      st->st_goal |= GOAL_TUNNEL;

    p = oakley_get_sa(&j, sin, st->st_goal, st);
    if (p == (u_char *) NULL)
    {
	free_state(st);
	free(buffer);
	if (idstore != (u_char *) NULL)
	  free(idstore);
	return;
    }

    bcopy(p, st->st_packet + i, j);

    /* Fix NextPayload field */
    isas = (struct isakmp_sa *) (st->st_packet + i);
    isas->isasa_np = ISAKMP_NEXT_NONCE;

    /* Also copy SA into state */
    st->st_sa = (u_char *) calloc(j, sizeof(u_char));
    if (st->st_sa == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_r1()");
    bcopy(p, st->st_sa, j);
    free(p);

    i += ntohs(isas->isasa_length);

    /* Nr */
    isag = (struct isakmp_generic *) (st->st_packet + i);
    isag->isag_length = htons(DEFAULTNONCESIZE +
			      sizeof(struct isakmp_generic));
    get_rnd_bytes((u_char *)isag + sizeof(struct isakmp_generic),
		  DEFAULTNONCESIZE);

    /* Copy Nr into state */
    st->st_nr = (u_char *) calloc(ntohs(isag->isag_length), sizeof(u_char));
    if (st->st_nr == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_r1()");

    bcopy(isag, st->st_nr, ntohs(isag->isag_length));

#ifdef DEBUG
    log(0, "Nr sent: ");
    for (k = 0; k < ntohs(isag->isag_length) - sizeof(struct isakmp_generic);
	 k++)
      fprintf(stderr, "%02x ", st->st_nr[sizeof(struct isakmp_generic) + k]);
    fprintf(stderr, "\n");
#endif

    /* Advance pointer, it now has the packet length */
    i += ntohs(isag->isag_length);

    /* Need to send back IDui/IDur */
    if (st->st_ouruser.sin_addr.s_addr != 0)
    {
	isag->isag_np = ISAKMP_NEXT_ID;
	bcopy(idstore, st->st_packet + i, idlength);
	i += idlength;
#ifdef DEBUG
	log(0, "appended %d bytes of IDui/IDur", idlength);
#endif
    }

    st->st_packet_len = i;
    isa->isa_length = htonl(i);

#ifdef DEBUG
    log(0, "packet length (before padding) is %d bytes", i);
#endif

    /* Compute HASH(2) */
    compute_hash(st, st->st_packet, st->st_packet_len, 1);

    /* Update state of exchange */
    st->st_state = OAKLEY_QUICK_R_2;

    /* Free received packet */
    free(buffer);

    /* Padding */
    switch (st->st_enc)
    {
	case OAKLEY_DES_CBC:
	case OAKLEY_3DES_CBC:
	    blocksize = DES_CBC_BLOCK_SIZE;
	    break;

	default:
	    exit_log("unknown/unsupported encryption algorithm %d specified",
		     st->st_enc);
    }

    k = (blocksize -
	 (st->st_packet_len - sizeof(struct isakmp_hdr)) % blocksize) %
	blocksize;
    st->st_packet_len += k;
    isa->isa_length = htonl(st->st_packet_len);

#ifdef DEBUG
    log(0, "added %d bytes of padding", k);
#endif

    /* Encrypt */
    switch (st->st_enc)
    {
	case OAKLEY_DES_CBC:
	    /* XXX Do weak key detection */
	    (void) des_set_key((des_cblock *)st->st_skeyid_e, des_cbc_keys);
	    des_cbc_encrypt((des_cblock *)(st->st_packet + sizeof(struct isakmp_hdr)),
			    (des_cblock *)(st->st_packet + sizeof(struct isakmp_hdr)),
			    st->st_packet_len - sizeof(struct isakmp_hdr),
			    des_cbc_keys, (des_cblock *)st->st_iv, 1);


#ifdef DEBUG
	    log(0, "encrypting using DES CBC");
#endif

	    /* Keep last block as IV */
	    if ((st->st_iv != (u_char *) NULL) &&
		(st->st_iv_len < DES_CBC_BLOCK_SIZE))
	    {
		free(st->st_iv);
		st->st_iv = (u_char *) NULL;
	    }

	    if (st->st_iv == (u_char *) NULL)
	    {
		st->st_iv = (u_char *) calloc(DES_CBC_BLOCK_SIZE,
					      sizeof(u_char));
		if (st->st_iv == (u_char *) NULL)
		  exit_log("calloc() failed in ipsecdoi_handle_quick_r1()");
	    }

	    bcopy(st->st_packet + st->st_packet_len - DES_CBC_BLOCK_SIZE ,
		  st->st_iv, DES_CBC_BLOCK_SIZE);
	    st->st_iv_len = DES_CBC_BLOCK_SIZE;

	    break;

	    /* XXX Support more */
	default:
	    exit_log("unsupported encryption algorithm %d", st->st_enc);
    }

#ifdef DEBUG
    log(0, "new IV: ");
    for (k = 0; k < st->st_iv_len; k++)
      fprintf(stderr, "%02x ", st->st_iv[k]);
    fprintf(stderr, "\n");
#endif

    /* Transmit */
    if (sendto(sock, st->st_packet, st->st_packet_len, 0, &sa, sizeof(sa))
	!= st->st_packet_len)
      log(1, "sendto() failed in ipsecdoi_handle_r2()");
#ifdef DEBUG
    else
      log(0, "transmitted %d bytes", st->st_packet_len);
#endif

    /* Schedule */
    event_schedule(EVENT_CLEANUP, EVENT_CLEANUP_DELAY, st, 0);
}

/*
 * Produce the new key material of Quick Mode.
 * XXX It doesn't handle expanding the keysize right now.
 */
static void
compute_keymat(struct state *st, int keylength)
{
    MD5_CTX md5ctx, md5ctx2;
    SHA1_CTX sha1ctx, sha1ctx2;
    u_char buf1[HMAC_BUFSIZE], buf2[HMAC_BUFSIZE];
    struct isakmp_generic *isag;
    int k, nilen, nrlen;

#ifdef DEBUG
    log(0, "in compute_keymat()");
#endif

    bzero(buf1, HMAC_BUFSIZE);
    bzero(buf2, HMAC_BUFSIZE);

    isag = (struct isakmp_generic *) st->st_ni;
    nilen = ntohs(isag->isag_length) - sizeof(struct isakmp_generic);

    isag = (struct isakmp_generic *) st->st_nr;
    nrlen = ntohs(isag->isag_length) - sizeof(struct isakmp_generic);

    if (st->st_prf == 0)
    {
	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		st->st_keymat_len = MD5_DIGEST_SIZE;
		st->st_peer_keymat_len = MD5_DIGEST_SIZE;
		break;

	    case OAKLEY_SHA:
		st->st_keymat_len = SHA1_DIGEST_SIZE;
		st->st_peer_keymat_len = SHA1_DIGEST_SIZE;
		break;

	    default:
		exit_log("unknown/unsupported hash algorithm %d specified",
			 st->st_hash);
	}
    }
    else
    {
	/* XXX Handle 3DES-CBC MAC */
    }

    st->st_keymat = (u_char *) calloc(st->st_keymat_len, sizeof(u_char));
    if (st->st_keymat == (u_char *) NULL)
      exit_log("calloc() failed in compute_keymat()");

    st->st_peer_keymat = (u_char *) calloc(st->st_peer_keymat_len,
					   sizeof(u_char));
    if (st->st_peer_keymat == (u_char *) NULL)
      exit_log("calloc() failed in compute_keymat()");

    /* Prepare the two pads for the HMAC */
    if (st->st_skeyid_d_len <= HMAC_BUFSIZE)
    {
	bcopy(st->st_skeyid_d, buf1, st->st_skeyid_d_len);
	bcopy(st->st_skeyid_d, buf2, st->st_skeyid_d_len);
    }
    else
    {
	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, st->st_skeyid_d, st->st_skeyid_d_len);
		MD5Final(buf1, &md5ctx);
		bcopy(buf1, buf2, MD5_DIGEST_SIZE);
		break;

	    case OAKLEY_SHA:
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, st->st_skeyid_d, st->st_skeyid_d_len);
		SHA1Final(buf1, &sha1ctx);
		bcopy(buf1, buf2, SHA1_DIGEST_SIZE);
		break;

	    default:
		exit_log("unknown/unsupported hash algorithm %d in compute_keymat()", st->st_hash);
	}
    }

    for (k = 0; k < HMAC_BUFSIZE; k++)
    {
	buf1[k] ^= HMAC_IPAD;
	buf2[k] ^= HMAC_OPAD;
    }

    if (st->st_prf == 0)
    {
	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		MD5Init(&md5ctx);
		MD5Init(&md5ctx2);
		MD5Update(&md5ctx, buf1, HMAC_BUFSIZE);
		MD5Update(&md5ctx2, buf1, HMAC_BUFSIZE);
		MD5Update(&md5ctx, &(st->st_protoid), sizeof(st->st_protoid));
		MD5Update(&md5ctx2, &(st->st_peer_protoid),
			  sizeof(st->st_protoid));
		MD5Update(&md5ctx, st->st_spi, st->st_spi_len);
		MD5Update(&md5ctx2, st->st_peer_spi, st->st_peer_spi_len);
		MD5Update(&md5ctx, st->st_ni + sizeof(struct isakmp_generic),
			  nilen);
		MD5Update(&md5ctx2, st->st_ni + sizeof(struct isakmp_generic),
			  nilen);
		MD5Update(&md5ctx, st->st_nr + sizeof(struct isakmp_generic),
			  nrlen);
		MD5Update(&md5ctx2, st->st_nr + sizeof(struct isakmp_generic),
			  nrlen);
		MD5Final(st->st_keymat, &md5ctx);
		MD5Final(st->st_peer_keymat, &md5ctx2);
		MD5Init(&md5ctx);
		MD5Init(&md5ctx2);
		MD5Update(&md5ctx, buf2, HMAC_BUFSIZE);
		MD5Update(&md5ctx2, buf2, HMAC_BUFSIZE);
		MD5Update(&md5ctx, st->st_keymat, MD5_DIGEST_SIZE);
		MD5Update(&md5ctx2, st->st_peer_keymat, MD5_DIGEST_SIZE);
		MD5Final(st->st_keymat, &md5ctx);
		MD5Final(st->st_peer_keymat, &md5ctx2);

		break;

	    case OAKLEY_SHA:
		SHA1Init(&sha1ctx);
		SHA1Init(&sha1ctx2);
		SHA1Update(&sha1ctx, buf1, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx2, buf1, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, &(st->st_protoid),
			   sizeof(st->st_protoid));
		SHA1Update(&sha1ctx2, &(st->st_peer_protoid),
			   sizeof(st->st_protoid));
		SHA1Update(&sha1ctx, st->st_spi, st->st_spi_len);
		SHA1Update(&sha1ctx2, st->st_peer_spi, st->st_peer_spi_len);
		SHA1Update(&sha1ctx, st->st_ni + sizeof(struct isakmp_generic),
			   nilen);
		SHA1Update(&sha1ctx2, st->st_ni +
			   sizeof(struct isakmp_generic), nilen);
		SHA1Update(&sha1ctx, st->st_nr + sizeof(struct isakmp_generic),
			   nrlen);
		SHA1Update(&sha1ctx2, st->st_nr +
			   sizeof(struct isakmp_generic), nrlen);
		SHA1Final(st->st_keymat, &sha1ctx);
		SHA1Final(st->st_peer_keymat, &sha1ctx2);
		SHA1Init(&sha1ctx);
		SHA1Init(&sha1ctx2);
		SHA1Update(&sha1ctx, buf2, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx2, buf2, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, st->st_keymat, SHA1_DIGEST_SIZE);
		SHA1Update(&sha1ctx2, st->st_peer_keymat, SHA1_DIGEST_SIZE);
		SHA1Final(st->st_keymat, &sha1ctx);
		SHA1Final(st->st_peer_keymat, &sha1ctx2);

		break;

	    default:
		exit_log("unknown/unsupported hash algorithm %d specified in compute_keymat()", st->st_hash);
	}
    }
    else
    {
	/* XXX Handle 3DES */
    }

#ifdef DEBUG
    fprintf(stdout, "KEYMAT computed: ");
    for (k = 0; k < st->st_keymat_len; k++)
      fprintf(stdout, "%02x ", st->st_keymat[k]);
    fprintf(stdout, "\n");
    fprintf(stdout, "Peer KEYMAT computed: ");
    for (k = 0; k < st->st_peer_keymat_len; k++)
      fprintf(stdout, "%02x ", st->st_peer_keymat[k]);
    fprintf(stdout, "\n");
#endif
}

/*
 * Compute HASH(3) in Quick Mode.
 */
static void
compute_hash_3(struct state *st, u_char *buffer)
{
    MD5_CTX md5ctx;
    SHA1_CTX sha1ctx;
    u_char *p, buf1[HMAC_BUFSIZE], buf2[HMAC_BUFSIZE];
    struct isakmp_generic *isag, *isag2;
    int k, nilen, nrlen;
    u_char c = 0;

    bzero(buf1, HMAC_BUFSIZE);
    bzero(buf2, HMAC_BUFSIZE);

    p = buffer + sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic);
    isag = (struct isakmp_generic *) (buffer + sizeof(struct isakmp_hdr));

    isag2 = (struct isakmp_generic *) st->st_ni;
    nilen = ntohs(isag2->isag_length) - sizeof(struct isakmp_generic);

    isag2 = (struct isakmp_generic *) st->st_nr;
    nrlen = ntohs(isag2->isag_length) - sizeof(struct isakmp_generic);

    /* Prepare the two pads for the HMAC */
    if (st->st_skeyid_a_len <= HMAC_BUFSIZE)
    {
	bcopy(st->st_skeyid_a, buf1, st->st_skeyid_a_len);
	bcopy(st->st_skeyid_a, buf2, st->st_skeyid_a_len);
    }
    else
    {
	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, st->st_skeyid_a, st->st_skeyid_a_len);
		MD5Final(buf1, &md5ctx);
		bcopy(buf1, buf2, MD5_DIGEST_SIZE);
		break;

	    case OAKLEY_SHA:
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, st->st_skeyid_a, st->st_skeyid_a_len);
		SHA1Final(buf1, &sha1ctx);
		bcopy(buf1, buf2, SHA1_DIGEST_SIZE);
		break;

	    default:
		exit_log("unknown/unsupported hash algorithm %d in compute_keymat()", st->st_hash);
	}
    }

    for (k = 0; k < HMAC_BUFSIZE; k++)
    {
	buf1[k] ^= HMAC_IPAD;
	buf2[k] ^= HMAC_OPAD;
    }

    if (st->st_prf == 0)
    {
	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf1, HMAC_BUFSIZE);
		MD5Update(&md5ctx, &c, sizeof(c));
		MD5Update(&md5ctx, (u_char *) &(st->st_msgid),
			  sizeof(st->st_msgid));
		MD5Update(&md5ctx, st->st_ni + sizeof(struct isakmp_generic),
			  nilen);
		MD5Update(&md5ctx, st->st_nr + sizeof(struct isakmp_generic),
			  nrlen);
		MD5Final(p, &md5ctx);
		MD5Init(&md5ctx);
		MD5Update(&md5ctx, buf2, HMAC_BUFSIZE);
		MD5Update(&md5ctx, p, MD5_DIGEST_SIZE);
		MD5Final(p, &md5ctx);

		break;

	    case OAKLEY_SHA:
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf1, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, &c, sizeof(c));
		SHA1Update(&sha1ctx, (u_char *) &(st->st_msgid),
			   sizeof(st->st_msgid));
		SHA1Update(&sha1ctx, st->st_ni + sizeof(struct isakmp_generic),
 			   nilen);
		SHA1Update(&sha1ctx, st->st_nr + sizeof(struct isakmp_generic),
			   nrlen);
		SHA1Final(p, &sha1ctx);
		SHA1Init(&sha1ctx);
		SHA1Update(&sha1ctx, buf2, HMAC_BUFSIZE);
		SHA1Update(&sha1ctx, p, SHA1_DIGEST_SIZE);
		SHA1Final(p, &sha1ctx);

		break;

	    default:
		exit_log("unknown/unsupported hash algorithm %d specified in compute_hash_3()", st->st_hash);
	}
    }
    else
    {
	/* XXX Handle 3DES */
    }

#ifdef DEBUG
    log(0, "HASH(3) computed: ");
    for (k = 0; k < ntohs(isag->isag_length) - sizeof(struct isakmp_generic);
	 k++)
      fprintf(stderr, "%02x ", p[k]);
    fprintf(stderr, "\n");
#endif
}

#if 0   /* This is no longer necessary; ipsecdoi_ipsec_sa_copy() does it. */
/*
 * Copy the attributes from an accepted proposal in the state. This
 * is only used by the Initiator. The proposal should alredy be in
 * the state.
 * XXX Doesn't handle multi-protocol proposals.
 */
void
ipsec_prop_to_state(struct state *st)
{
    struct isakmp_sa *isasa;
    struct isakmp_proposal *isap;
    struct isakmp_transform *isat;
    struct isakmp_attribute *isaa, isab;
    int i, k;
    int salifetype, authlifetype, enclifetype;

    isasa = (struct isakmp_sa *) st->st_proposal;
    i = sizeof(struct isakmp_sa) + IPSEC_DOI_SITUATION_LENGTH;
    isap = (struct isakmp_proposal *) (st->st_proposal + i);
    i += sizeof(struct isakmp_proposal);

#ifdef DEBUG
    log(0, "SPI accepted: ");
    for (k = 0; k < isap->isap_spisize; k++)
      fprintf(stderr, "%02x ", st->st_proposal[i + k]);
    fprintf(stderr, "\n");
#endif

    st->st_spi = (u_char *) calloc(isap->isap_spisize, sizeof(u_char));
    if (st->st_spi == (u_char *) NULL)
      exit_log("calloc() failed in ipsec_prop_to_state()");

    bcopy(st->st_proposal + i, st->st_spi, isap->isap_spisize);
    st->st_spi_len = isap->isap_spisize;

    st->st_protoid = isap->isap_protoid;
    i += isap->isap_spisize;

#ifdef DEBUG
    log(0, "protocol ID %d", st->st_protoid);
#endif

    isat = (struct isakmp_transform *) (st->st_proposal + i);
    st->st_transid = isat->isat_transid;

#ifdef DEBUG
    log(0, "transform ID %d", st->st_transid);
#endif

    i += sizeof(struct isakmp_transform);
    k = 0;
    salifetype = 0;
    authlifetype = 0;
    enclifetype = 0;

#ifdef DEBUG
    log(0, "i = %d, length = %d", i, ntohs(isasa->isasa_length));
#endif

    while (i < ntohs(isasa->isasa_length))
    {
	isaa = (struct isakmp_attribute *) (st->st_proposal + i);
	isab.isaat_type = ntohs(isaa->isaat_type);
	isab.isaat_lv = ntohs(isaa->isaat_lv);

	if (isab.isaat_af)
	  k = sizeof(struct isakmp_attribute);
	else
	  k = isab.isaat_lv;

	/* XXX Should handle TLVs too in the switch */

	switch (isab.isaat_rtype)
	{
		case SA_LIFE_TYPE:
#ifdef DEBUG
		    log(0, "SA life type %d", isab.isaat_lv);
#endif
		    salifetype = isab.isaat_lv;
		    break;

		case SA_LIFE_DURATION:
#ifdef DEBUG
		    log(0, "SA life duration %u", isab.isaat_lv);
#endif
		    switch (salifetype)
		    {
			case SA_LIFE_TYPE_SECONDS:
			    st->st_expire = isab.isaat_lv;
			    break;

			case SA_LIFE_TYPE_KBYTES:
			    st->st_sa_duration = isab.isaat_lv;
			    break;
		    }

		    break;

		case AUTH_KEY_LIFE_TYPE:
#ifdef DEBUG
		    log(0, "auth key life type %d", isab.isaat_lv);
#endif
		    authlifetype = isab.isaat_lv;
		    break;

		case AUTH_KEY_LIFE_DURATION:
#ifdef DEBUG
		    log(0, "auth key life duration %u", isab.isaat_lv);
#endif
		    switch (authlifetype)
		    {
			case SA_LIFE_TYPE_SECONDS:
			    st->st_authkey_duration = isab.isaat_lv;
			    break;

			case SA_LIFE_TYPE_KBYTES:
			    st->st_authkey_k_duration = isab.isaat_lv;
			    break;
		    }

		    break;

		case ENC_KEY_LIFE_TYPE:
#ifdef DEBUG
		    log(0, "enc key life type %d", isab.isaat_lv);
#endif
		    enclifetype = isab.isaat_lv;
		    break;

		case ENC_KEY_LIFE_DURATION:
#ifdef DEBUG
		    log(0, "enc key life duration %d", isab.isaat_lv);
#endif
		    switch (enclifetype)
		    {
			case SA_LIFE_TYPE_SECONDS:
			    st->st_enckey_duration = isab.isaat_lv;
			    break;

			case SA_LIFE_TYPE_KBYTES:
			    st->st_enckey_k_duration = isab.isaat_lv;
			    break;
		    }

		    break;

		case GROUP_DESCRIPTION:
#ifdef DEBUG
		    log(0, "group description %d", isab.isaat_lv);
#endif
		    st->st_group = isab.isaat_lv;
		    break;

		case AUTH_ALGORITHM:
#ifdef DEBUG
		    log(0, "AUTH algorithm %d", isab.isaat_lv);
#endif
		    st->st_hmac_algorithm = isab.isaat_lv;
		    break;

		case CA_DISTINGUISHED_NAME:
#ifdef DEBUG
		    log(0, "CA distinguished name type %d",
			isab.isaat_lv);
#endif
		    st->st_ca_name = isab.isaat_lv;
		    break;

		case ENCAPSULATION_MODE:
#ifdef DEBUG
		    log(0, "encapsulation mode %d", isab.isaat_lv);
#endif
		    st->st_encapsulation_mode = isab.isaat_lv;
		    break;
	}

	i += k;
    }

    /* Install some default values if not already negotiated */
    if (st->st_expire == 0)
#ifdef DEBUG
    {
	log(0, "default SA lifetime %u seconds", SA_LIFE_DURATION_DEFAULT);
#endif
	st->st_expire = SA_LIFE_DURATION_DEFAULT;
#ifdef DEBUG
    }
#endif

    if (st->st_authkey_duration == 0)
#ifdef DEBUG
    {
	log(0, "default auth key lifetime %u seconds",
	    SA_LIFE_DURATION_DEFAULT);
#endif
	    st->st_authkey_duration = SA_LIFE_DURATION_DEFAULT;
#ifdef DEBUG
    }
#endif

    if (st->st_enckey_duration == 0)
#ifdef DEBUG
    {
	log(0, "default enc key lifetime %u seconds",
	    SA_LIFE_DURATION_DEFAULT);
#endif
	st->st_enckey_duration = SA_LIFE_DURATION_DEFAULT;
#ifdef DEBUG
    }
#endif

    if (st->st_ca_name == 0)
#ifdef DEBUG
    {
	log(0, "default CA distinguished name type %d",
	    CA_DISTINGUISHED_NAME_DNSSEC);
#endif
	st->st_ca_name = CA_DISTINGUISHED_NAME_DNSSEC;
#ifdef DEBUG
    }
#endif
}

#endif

/*
 * Handle (the single) message from Responder in Quick Mode.
 */
void
ipsecdoi_handle_quick_i1(int sock, u_char *buffer, int length,
			 struct sockaddr sa, struct state *st, int kernsock)
{
    struct isakmp_hdr *isa;
    struct isakmp_generic *isag, *isag2;
    struct isakmp_sa *isasa;
    u_char *hashval;
    int i, k, blocksize, hashlen;
    des_key_schedule des_cbc_keys;
    u_int32_t *iaddr;
    struct identity *id;

    isa = (struct isakmp_hdr *) buffer;

    if (length <= sizeof(struct isakmp_hdr) + sizeof(struct isakmp_sa) +
	2 * sizeof(struct isakmp_generic))
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    isag = (struct isakmp_generic *) (buffer + sizeof(struct isakmp_hdr));

    if ((isa->isa_np != ISAKMP_NEXT_HASH) ||
	(isag->isag_np != ISAKMP_NEXT_SA))
    {
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    i = sizeof(struct isakmp_hdr);
    if (i + ntohs(isag->isag_length) >= length)
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    i += sizeof(struct isakmp_generic);

    /* Check HASH(2) */
    k = ntohs(isag->isag_length) - sizeof(struct isakmp_generic);
    hashval = (u_char *) calloc(k, sizeof(u_char));
    if (hashval == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_i1()");
    bcopy(buffer + i, hashval, k);

    compute_hash(st, buffer, length, 1);

#ifdef DEBUG
    log(0, "HASH(2) received: ");
    for (k = 0; k < ntohs(isag->isag_length) - sizeof(struct isakmp_generic);
	 k++)
      fprintf(stderr, "%02x ", hashval[k]);
    fprintf(stderr, "\n");
#endif

    if (bcmp(buffer + i, hashval,
	     ntohs(isag->isag_length) - sizeof(struct isakmp_generic)))
    {
	log(0, "HASH(2) authentication failed");
	free(hashval);
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    free(hashval);

#ifdef DEBUG
    log(0, "HASH(2) verified");
#endif

    i = sizeof(struct isakmp_hdr) + ntohs(isag->isag_length);

    /* SA */
    isasa = (struct isakmp_sa *) (buffer + i);

    if (ntohs(isasa->isasa_length) + i >= length)
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    if (isasa->isasa_np != ISAKMP_NEXT_NONCE)
    {
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    /* Copy the SA we got into the state, while verifying it */
    if (ipsecdoi_ipsec_sa_copy(buffer, length, sa, st) == 0)
    {
	free(buffer);
	free_state(st);
	log(0, "no acceptable proposal from %s, port %d",
	    get_address(sa), get_port(sa));
	/* XXX Could send notification back */
	return;
    }

    i += ntohs(isasa->isasa_length);

    /* Nr */
    isag = (struct isakmp_generic *) (buffer + i);

    if (ntohs(isag->isag_length) + i > length)
    {
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    isag2 = isag;	/* Keep this around */

    /* Handle additional payloads */
    switch (isag->isag_np)
    {
	case ISAKMP_NEXT_ID:
#ifdef DEBUG
	    log(0, "IDui/IDur present");
#endif

	    if (st->st_ouruser.sin_addr.s_addr == 0)
	    {
		log(0, "IDui/IDur present, when they shouldn't have been, from %s, port %d", get_address(sa), get_port(sa));
		free(buffer);
		/* XXX Could send notification back */
		return;
	    }

	    isag = (struct isakmp_generic *) ((u_char *) isag +
 					      ntohs(isag->isag_length));
	    if (isag->isag_np != ISAKMP_NEXT_ID)
	    {
		log(0, "malformed packet from %s, port %d", get_address(sa),
	    	    get_port(sa));
		free(buffer);
		/* XXX Could send notification back */
		return;
	    }

	    /* IDui */
	    id = (struct identity *) (isag + 1);
	    if ((id->id_protoid != st->st_ouruserprotoid) ||
		(id->id_port != st->st_ouruserport))
	    {
		log(0,
		    "our user protocol or port doesn't match returned values");
		free(buffer);
		/* XXX Could send notification back */
		return;
	    }

	    /* XXX support more */
	    switch (id->id_type)
	    {
		case ID_IPV4_ADDR:
		    iaddr = (u_int32_t *) (id + 1);
		    if ((st->st_ouruser.sin_addr.s_addr != *iaddr) ||
			(st->st_ournetmask.sin_addr.s_addr = 0xffffffff))
		    {
			log(0, "our user IPv4 address not verified");
			free(buffer);
			/* XXX Could send notification back */
			return;
		    }

#ifdef DEBUG
		    log(0, "our user IPv4 address verified");
#endif
		    break;

		case ID_IPV4_ADDR_SUBNET:
		    iaddr = (u_int32_t *) (id + 1);

		    if ((st->st_ouruser.sin_addr.s_addr != *(iaddr++)) ||
			(st->st_ournetmask.sin_addr.s_addr != *iaddr))
		    {
			log(0, "our user IPv4 subnet not verified");
			free(buffer);
			/* XXX Could send notification back */
			return;
		    }

#ifdef DEBUG
		    log(0, "our user IPv4 subnet verified");
#endif
		    break;

		default:
		    log(0, "bad id type %d from %s, port %d", id->id_type,
			get_address(sa), get_port(sa));
		    free(buffer);
		    /* XXX Could send notification back */
		    return;
	    }

	    /* IDur */
	    isag = (struct isakmp_generic *) ((u_char *) isag +
					      ntohs(isag->isag_length));
	    if (isag->isag_np != ISAKMP_NEXT_NONE)
	    {
                log(0, "malformed packet from %s, port %d", get_address(sa),
                    get_port(sa));
                free(buffer);
                /* XXX Could send notification back */
                return;
            }

            id = (struct identity *) (isag + 1);
	    if ((st->st_peeruserport != id->id_port) ||
		(st->st_peeruserprotoid = id->id_protoid))
	    {
		log(0,
		    "peer user protocol or port doesn't match returned values");
		free(buffer);
		/* XXX Could send notification back */
		return;
	    }

            /* XXX support more */
            switch (id->id_type)
            {
                case ID_IPV4_ADDR:
                    iaddr = (u_int32_t *) (id + 1);

                    if ((st->st_peeruser.sin_addr.s_addr != *iaddr) ||
			(st->st_peernetmask.sin_addr.s_addr != 0xffffffff))
		    {
			log(0, "peer user IPv4 address not verified");
			free(buffer);
			/* XXX Could send notification back */
			return;
		    }

#ifdef DEBUG
                    log(0, "peer user IPv4 address verified");
#endif
                    break;

                case ID_IPV4_ADDR_SUBNET:
                    iaddr = (u_int32_t *) (id + 1);
                    if ((st->st_peeruser.sin_addr.s_addr != *(iaddr++)) ||
			(st->st_peernetmask.sin_addr.s_addr != *iaddr))
		    {
			log(0, "peer user IPv4 subnet address not verified");
			free(buffer);
			/* XXX Could send notification back */
			return;
		    }

#ifdef DEBUG
                    log(0, "peer user IPv4 subnet address verified");
#endif
                    break;

                default:
                    log(0, "bad id type %d from %s, port %d", id->id_type,
                        get_address(sa), get_port(sa));
                    free(buffer);
                    /* XXX Could send notification back */
                    return;
            }

	    break;

	case ISAKMP_NEXT_NONE:
	    break;

	case ISAKMP_NEXT_KE:
	    /* XXX Unsupported now, just fall through */
	default:
	    log(0,
		"unsupported payload received in Quick Mode from %s, port %d",
	     	get_address(sa), get_port(sa));
	    free(buffer);
	    /* XXX Could send notification back */
	    return;
    }

    /* Copy Nr into the state */
    st->st_nr = (u_char *) calloc(ntohs(isag2->isag_length), sizeof(u_char));
    if (st->st_nr == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_r1()");
    bcopy(isag2, st->st_nr, ntohs(isag2->isag_length));

#ifdef DEBUG
    log(0, "Nr received: ");
    for (k = 0; k < ntohs(isag2->isag_length) - sizeof(struct isakmp_generic);
	 k++)
      fprintf(stderr, "%02x ", st->st_nr[k + sizeof(struct isakmp_generic)]);
    fprintf(stderr, "\n");
#endif

    /* Copy accepted proposal into the state */
    st->st_proposal = (u_char *) calloc(ntohs(isasa->isasa_length),
					sizeof(u_char));
    if (st->st_proposal == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_i1()");
    bcopy(isasa, st->st_proposal, ntohs(isasa->isasa_length));

    /* XXX Check that the proposal is one of those we sent originally */

#if 0
    /* Copy the proposal attributes in the state object */
    ipsec_prop_to_state(st);
#endif

    /* Get rid of old packet */
    free(st->st_packet);

    /* Delete old pending event */
    delete_event(st);

    /* New packet */
    if (st->st_prf == 0)
    {
	switch (st->st_hash)
	{
	    case OAKLEY_MD5:
		hashlen = MD5_DIGEST_SIZE;
		break;

	    case OAKLEY_SHA:
		hashlen = SHA1_DIGEST_SIZE;
		break;

	    default:
		exit_log("unknown/unsupported hash algorithm %d specified",
			 st->st_hash);
	}
    }
    else
    {
	switch (st->st_prf)
	{
	    case OAKLEY_PRF_3DES_CBC_MAC:
		hashlen = DES_CBC_BLOCK_SIZE;
		break;
	    default:
		exit_log("unknown/unsupported PRF algorithm %d specified",
			 st->st_prf);
	}
    }

    st->st_packet_len = sizeof(struct isakmp_hdr) +
			sizeof(struct isakmp_generic) + hashlen;
    st->st_packet = (u_char *) calloc(st->st_packet_len, sizeof(u_char));
    if (st->st_packet == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_i1()");

    bcopy(buffer, st->st_packet, sizeof(struct isakmp_hdr));

    isa = (struct isakmp_hdr *) st->st_packet;
    isa->isa_np = ISAKMP_NEXT_HASH;

    /* HASH(3) */
    isag = (struct isakmp_generic *) (st->st_packet +
				      sizeof(struct isakmp_hdr));
    isag->isag_length = htons(hashlen + sizeof(struct isakmp_generic));

    /* Compute HASH(3) */
    compute_hash_3(st, st->st_packet);

    /* Free old packets */
    free(buffer);

    /* Advance state */
    st->st_state = OAKLEY_QUICK_I_2;

    /* Padding */
    switch (st->st_enc)
    {
	case OAKLEY_DES_CBC:
	case OAKLEY_3DES_CBC:
	    blocksize = DES_CBC_BLOCK_SIZE;
	    break;

	default:
	    exit_log("unknown/unsupported encryption algorithm %d specified",
		     st->st_enc);
    }

    k = (blocksize -
	 (st->st_packet_len - sizeof(struct isakmp_hdr)) % blocksize) %
	blocksize;
    st->st_packet_len += k;
    isa->isa_length = htonl(st->st_packet_len);

#ifdef DEBUG
    log(0, "added %d bytes of padding", k);
#endif

    /* Encrypt */
    switch (st->st_enc)
    {
	case OAKLEY_DES_CBC:
	    /* XXX Do weak key detection */
	    (void) des_set_key((des_cblock *)st->st_skeyid_e, des_cbc_keys);
	    des_cbc_encrypt((des_cblock *)(st->st_packet + sizeof(struct isakmp_hdr)),
			    (des_cblock *)(st->st_packet + sizeof(struct isakmp_hdr)),
			    st->st_packet_len - sizeof(struct isakmp_hdr),
			    des_cbc_keys, (des_cblock *)st->st_iv, 1);


#ifdef DEBUG
	    log(0, "encrypting using DES CBC");
#endif

	    /* Keep last block as IV */
	    if ((st->st_iv != (u_char *) NULL) &&
		(st->st_iv_len < DES_CBC_BLOCK_SIZE))
	    {
		free(st->st_iv);
		st->st_iv = (u_char *) NULL;
	    }

	    if (st->st_iv == (u_char *) NULL)
	    {
		st->st_iv = (u_char *) calloc(DES_CBC_BLOCK_SIZE,
					      sizeof(u_char));
		if (st->st_iv == (u_char *) NULL)
		  exit_log("calloc() failed in ipsecdoi_handle_quick_i1()");
	    }

	    bcopy(st->st_packet + st->st_packet_len - DES_CBC_BLOCK_SIZE ,
		  st->st_iv, DES_CBC_BLOCK_SIZE);
	    st->st_iv_len = DES_CBC_BLOCK_SIZE;

	    break;

	    /* XXX Support more */
	default:
	    exit_log("unsupported encryption algorithm %d", st->st_enc);
    }

#ifdef DEBUG
    log(0, "new IV: ");
    for (k = 0; k < st->st_iv_len; k++)
      fprintf(stderr, "%02x ", st->st_iv[k]);
    fprintf(stderr, "\n");
#endif

    /* Transmit */
    if (sendto(sock, st->st_packet, st->st_packet_len, 0, &sa, sizeof(sa))
	!= st->st_packet_len)
      log(1, "sendto() failed in ipsecdoi_handle_i1()");
#ifdef DEBUG
    else
      log(0, "transmitted %d bytes", st->st_packet_len);
#endif

    /* Derive new keying material */
    compute_keymat(st, 0);

    /* Schedule SA expiration */
    if (st->st_expire != 0)
      event_schedule(EVENT_SA_EXPIRE, st->st_expire, st, 0);
    else
      event_schedule(EVENT_SA_EXPIRE, DEFAULT_SA_EXPIRE, st, 0);

    /*
     * XXX Unless the commit bit is set, we need to send a message
     * XXX to the kernel to establish the new SA.
     */

#ifdef JI
     setup_route(st);
     setup_sa(st);
#endif
}

/*
 * Handle last message of Quick Mode.
 */
void
ipsecdoi_handle_quick_r2(int sock, u_char *buffer, int length,
			 struct sockaddr sa, struct state *st,
			 int kernsock)
{
    struct isakmp_hdr *isa;
    struct isakmp_generic *isag;
    int i, k, j;
    u_char *hashval;

    isa = (struct isakmp_hdr *) buffer;

    /* Standard checks */
    if (length <= sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic))
    {
	log(0, "too short packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    if (isa->isa_np != ISAKMP_NEXT_HASH)
    {
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    isag = (struct isakmp_generic *) (buffer + sizeof(struct isakmp_hdr));

    if ((ntohs(isag->isag_length) + sizeof(struct isakmp_hdr) != length) ||
	(isag->isag_np != ISAKMP_NEXT_NONE))
    {
	log(0, "malformed packet from %s, port %d", get_address(sa),
	    get_port(sa));
	free(buffer);
	/* XXX Could send notification back */
	return;
    }

    /* Check HASH(3) */
    i = sizeof(struct isakmp_hdr) + sizeof(struct isakmp_generic);
    k = ntohs(isag->isag_length) - sizeof(struct isakmp_generic);
    hashval = (u_char *) calloc(k, sizeof(u_char));
    if (hashval == (u_char *) NULL)
      exit_log("calloc() failed in ipsecdoi_handle_quick_r2()");
    bcopy(buffer + i, hashval, k);

    compute_hash_3(st, buffer);

#ifdef DEBUG
    log(0, "HASH(3) received: ");
    for (j = 0; j < k; j++)
      fprintf(stderr, "%02x ", hashval[j]);
    fprintf(stderr, "\n");
#endif

    if (bcmp(buffer + i, hashval, k))
    {
	log(0, "HASH(3) authentication failed with %s, port %d",
	    get_address(sa), get_port(sa));
	free(buffer);
	free(hashval);
	/* XXX Could send notification back */
	return;
    }

    free(hashval);

#ifdef DEBUG
    log(0, "HASH(3) authentication succeeded");
#endif

    /* Delete old event */
    delete_event(st);

    /* Advance state */
    st->st_state = OAKLEY_QUICK_R_3;

    /* Free received packet */
    free(buffer);

    /* Schedule SA expiration */
    if (st->st_expire != 0)
      event_schedule(EVENT_SA_EXPIRE, st->st_expire, st, 0);
    else
      event_schedule(EVENT_SA_EXPIRE, DEFAULT_SA_EXPIRE, st, 0);

    /* Free old packet */
    free(st->st_packet);

    /* Derive new keying material */
    compute_keymat(st, 0);

    /*
     * XXX Unless the commit bit is set, we need to send a message
     * XXX to the kernel to establish the new SA.
     */

#ifdef JI
     setup_route(st);
     setup_sa(st);
#endif
}
