/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>

#include "log.h"

/*
 * This file defines error logging functions
 */

void
init_log(void)
{
#ifndef DEBUG
    openlog("Pluto", LOG_CONS | LOG_NDELAY | LOG_PID, LOG_AUTHPRIV);
#else
    setbuf(stderr, NULL);
    setbuf(stdout, NULL);
#endif
}

static void
close_log(void)
{
#ifndef DEBUG
    closelog();
#endif
}

/*
 * If the first argument is true, then this is an error, otherwise
 * it's a warning. It takes up to three additional arguments (which
 * will be passed to syslog/printf verbatim.
 */

#include <stdarg.h>

void
log(int error, char *message, ...)
{
    va_list args;
    int e = errno;  /* capture before we change it (may be too late) */
    char m[1024];

    va_start(args, message);
    vsnprintf(m, sizeof(m), message, args);
    va_end(args);

#ifndef DEBUG
    syslog(error? LOG_ERR : LOG_WARNING, "%s", m);
    if (error && e != 0)
	syslog(LOG_ERR, "errno %d: %s", e, sys_errlist[e]);
#else
    fprintf(stderr, "%s%s\n", error? "Error: " : "", m);
    if (error && e != 0)
	fprintf(stderr, "Errno %d: %s\n", e, sys_errlist[e]);
#endif
}

void
exit_log(char *message, ...)
{
    va_list args;
    int e = errno;  /* capture before we change it (may be too late) */
    char m[1024];

    va_start(args, message);
    vsnprintf(m, sizeof(m), message, args);
    va_end(args);

#ifndef DEBUG
    syslog(LOG_ERR, "%s", m);
    if (e != 0)
	syslog(LOG_ERR, "errno %d: %s", e, sys_errlist[e]);
#else
    fprintf(stderr, "Error: %s\n", m);
    if (e != 0)
	fprintf(stderr, "errno %d: %s\n", e, sys_errlist[e]);
#endif
    close_log();
    exit(-1);
}


/*
 * This routine returns a user readable form of an address contained in a
 * sockaddr structure. The return value CAN be a statically allocated
 * object (as is the case with inet_ntoa()).
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

char *
get_address_in(struct sockaddr_in sin)
{
    static char mess[50];

    switch (sin.sin_family)
    {
	case AF_INET:
	    return inet_ntoa(sin.sin_addr);

	default:
	    snprintf(mess, sizeof(mess), "(UNEXPECTED address family %d)",
		sin.sin_family);
	    return mess;
    }
}

char *
get_address(struct sockaddr sa)
{
    static char mess[50];
    struct sockaddr_in sin;

    switch (sa.sa_family)
    {
	case AF_INET:
	    bcopy(&sa, &sin, sizeof(sa));
	    return get_address_in(sin);

	default:
	    snprintf(mess, sizeof(mess), "(unknown address family %d)",
		sin.sin_family);
	    return mess;
    }
}

/*
 * Return a port (if applicable), from a struct sockaddr. If not applicable,
 * return -1.
 */

int
get_port_in(struct sockaddr_in sin)
{
    switch (sin.sin_family)
    {
	case AF_INET:
	    return ntohs(sin.sin_port);

	default:
	    return -1;
    }
}

int
get_port(struct sockaddr sa)
{
    struct sockaddr_in sin;

    switch (sa.sa_family)
    {
	case AF_INET:
	    bcopy(&sa, &sin, sizeof(sa));
	    return get_port_in(sin);

	default:
	    return -1;
    }
}
