/*
 * This program is an implementation of the ISAKMP Internet Standard.
 * Copyright (C) 1997 Angelos D. Keromytis.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * This code was written while the author was in Greece, in May/June
 * 1997.
 *
 * You may contact the author by:
 *   e-mail: angelos@dsl.cis.upenn.edu
 *  US-mail: Angelos D. Keromytis
 *           Distributed Systems Lab
 *           Computer and Information Science Department
 *           University of Pennsylvania
 *           Moore Building
 *           200 South 33rd Street
 *           Philadelphia, PA 19104
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>

#include "constants.h"
#include "state.h"
#include "defs.h"
#include "log.h"
#include "rnd.h"
#include "timer.h"

/*
 * This file has the functions that handle the
 * state hash table and the messageid list.
 */

#define STATE_TABLE_SIZE 32

static struct state *statetable[STATE_TABLE_SIZE];
static struct messageid *msgidlist;

#define SA_EQUAL(x, y)   ((x).sa_family == (y).sa_family && \
    bcmp(&((x).sa_data), &((y).sa_data), FULL_INET_ADDRESS_SIZE) == 0)

/*
 * Initialize the modulo and generator.
 */
void
init_nums(void)
{
    if (mpz_init_set_str(&groupmodulo, GROUPDEFAULT, 16) != 0)
      exit_log("mpz_init_set_str() failed in init_nums()");

    if (mpz_init_set_str(&groupgenerator, DEFAULTGENERATOR, 10) != 0)
      exit_log("mpz_init_set_str() failed in init_nums()");
}

/*
 * Global variables: had to go somewhere, might as well be this file.
 */

u_int16_t our_port;
u_char    secret_of_the_day[SECRET_VALUE_LENGTH];

MP_INT groupmodulo;
MP_INT groupgenerator;

void
init_vars(void)
{
    if (our_port == 0)
      our_port = PORT;

    /*
     * Generate the secret value for responder cookies, and
     * schedule an event for refresh.
     */
    get_rnd_bytes(secret_of_the_day, SECRET_VALUE_LENGTH);
    event_schedule(EVENT_REINIT_SECRET, EVENT_REINIT_SECRET_DELAY, NULL, 0);
}

/*
 * Insert a new message-id-keeping-track structure.
 */
struct messageid *
insert_messageid(struct sockaddr sa)
{
    struct messageid *mid;

    mid = (struct messageid *) calloc(1, sizeof(struct messageid));
    if (mid == (struct messageid *) NULL)
      exit_log("calloc() failed in insert_messageid()");

#ifdef DEBUG
    log(0, "inserting messageid structure for %s, port %d",
	get_address(sa), get_port(sa));
#endif

    mid->md_peer = sa;
    mid->md_msgid = 1;
    mid->md_next = msgidlist;
    msgidlist = mid;
    return mid;
}

/*
 * Update a messageid structure. If one doesn't exist, create it.
 */
void
update_messageid(struct sockaddr *sa, u_int32_t msgid)
{
    struct messageid *mid;

    for (mid = msgidlist; mid != (struct messageid *) NULL; mid = mid->md_next)
      if (SA_EQUAL(mid->md_peer, *sa))
	if (mid->md_msgid > msgid)
	  return;
	else
	{
#ifdef DEBUG
	    log(0, "updating MSG-ID for %s from %d to %d",
		get_address(*sa), mid->md_msgid, msgid);
#endif
	    mid->md_msgid = msgid;
	    return;
	}

    /* Not found, create one. */
    mid = insert_messageid(*sa);
    mid->md_msgid = msgid;
}

/*
 * Find a message-id-keeping-track structure, and return the
 * next msgid. If non, return 0.
 */
u_int32_t
find_messageid(struct sockaddr sa)
{
    struct messageid *mid;

    for (mid = msgidlist; mid != (struct messageid *) NULL; mid = mid->md_next)
      if (SA_EQUAL(mid->md_peer, sa))
        return ++(mid->md_msgid);

#ifdef DEBUG
    log(0, "find_messageid(): search failed, no structure for %s, port %d",
	get_address(sa), get_port(sa));
#endif

    return 0;
}

static u_int
state_hash(u_char *icookie, u_char *rcookie, struct sockaddr sa)
{
    u_int i, j;

#ifdef DEBUG2
    log(0, "ICOOKIE: ");
    for (i = 0; i < COOKIE_SIZE; i++)
      fprintf(stderr, "%02x ", icookie[i]);
    fprintf(stderr, "\n");
    log(0, "RCOOKIE: ");
    for (i = 0; i < COOKIE_SIZE; i++)
      fprintf(stderr, "%02x ", rcookie[i]);
    fprintf(stderr, "\n");
    log(0, "sockaddr: ");
    for (i = 0; i < 4; i++)
      fprintf(stderr, "%02x ", sa.sa_data[i]);
    fprintf(stderr, "\n");
#endif

    for (i = 0, j = 0; j < COOKIE_SIZE; j++)
      i += icookie[j] + rcookie[j];

    i += sa.sa_family;

    for (j = 0; j < FULL_INET_ADDRESS_SIZE; j++)
      i += sa.sa_data[j];

#ifdef DEBUG
    log(0, "state hash entry %d", i % STATE_TABLE_SIZE);
#endif

    return i % STATE_TABLE_SIZE;
}

/*
 * Get a state object.
 */
struct state *
get_state(void)
{
    struct state *st;

    st = (struct state *) calloc(1, sizeof(struct state));
    if (st == (struct state *) NULL)
      exit_log("calloc() failed in get_state()");

    return st;
}

/*
 * Initialize the state table.
 */
void
init_state(void)
{
    int i;

    for (i = 0; i < STATE_TABLE_SIZE; i++)
      statetable[i] = (struct state *) NULL;

    msgidlist = (struct messageid *) NULL;
}

/*
 * Insert a state object in the hash table. The object is inserted
 * at the begining of list.
 */
void
insert_state(struct state *st)
{
    int i;

    i = state_hash(st->st_icookie, st->st_rcookie, st->st_peer);

    st->st_prev = (struct state *) NULL;
    st->st_next = statetable[i];
    statetable[i] = st;
    if (st->st_next != (struct state *) NULL)
      st->st_next->st_prev = st;
}

/*
 * Delete a state object from the hash table, but don't free it
 */
void
delete_state(struct state *st)
{
    int i;

    /* If it's not the first in the list */

    if (st->st_prev != (struct state *) NULL)
    {
	st->st_prev->st_next = st->st_next;

	if (st->st_next != (struct state *) NULL)
	  st->st_next->st_prev = st->st_prev;

	return;
    }

    i = state_hash(st->st_icookie, st->st_rcookie, st->st_peer);
    statetable[i] = st->st_next;
    if (st->st_next != (struct state *) NULL)
      st->st_next->st_prev = (struct state *) NULL;
}

/*
 * Free a state object
 */
void
free_state(struct state *st)
{
    if ((st->st_packet != (u_char *) NULL) &&
	(st->st_packet_len != 0))
      free(st->st_packet);

    if (st->st_gi_in_use)
      mpz_clear(&(st->st_gi));

    if (st->st_gr_in_use)
      mpz_clear(&(st->st_gr));

    if (st->st_sec_in_use)
      mpz_clear(&(st->st_sec));

    if (st->st_shared_in_use)
      mpz_clear(&(st->st_shared));

#   define freeany(p) { if ((p) == NULL) free(p); }
    freeany(st->st_proposal);
    freeany(st->st_sa);
    freeany(st->st_ni);
    freeany(st->st_nr);
    freeany(st->st_skeyid);
    freeany(st->st_skeyid_d);
    freeany(st->st_skeyid_a);
    freeany(st->st_skeyid_e);
    freeany(st->st_myidentity);
    freeany(st->st_peeridentity);
    freeany(st->st_iv);
    freeany(st->st_lastblock);
    freeany(st->st_spi);
    freeany(st->st_keymat);
    freeany(st->st_peer_spi);
    freeany(st->st_peer_keymat);
#   undef freeany

    free(st);
}

/*
 * Duplicate a state object, for Phase 2.
 */
struct state *
duplicate_state(struct state *st)
{
    struct state *nst;

    nst = get_state();

    nst->st_phase1 = st;

    bcopy(st->st_icookie, nst->st_icookie, COOKIE_SIZE);
    bcopy(st->st_rcookie, nst->st_rcookie, COOKIE_SIZE);
    nst->st_peer = st->st_peer;

    nst->st_doi = st->st_doi;
    nst->st_situation = st->st_situation;

#   define clone_fld(fld, fld_len) { \
	nst->fld = (u_char *) malloc(st->fld_len); \
	if (nst->fld == (u_char *) NULL) \
	  exit_log("malloc() failed in duplicate_state()"); \
	\
	bcopy(st->fld, nst->fld, st->fld_len); \
	nst->fld_len = st->fld_len; \
    }

    clone_fld(st_skeyid_d, st_skeyid_d_len);	/* Copy SKEYID_D */
    clone_fld(st_skeyid_e, st_skeyid_e_len);	/* Copy SKEYID_E */
    clone_fld(st_skeyid_a, st_skeyid_a_len);	/* Copy SKEYID_A */
    clone_fld(st_myidentity, st_myidentity_len);    /* Copy myidentity */
#   undef clone_fld

    /* Copy goal */
    nst->st_goal = st->st_goal;

    nst->st_hash = st->st_hash;
    nst->st_enc = st->st_enc;
    nst->st_prf = st->st_prf;
    nst->st_auth = st->st_auth;

    return nst;
}

/*
 * Find a state object.
 */
struct state *
find_full_state(u_char *icookie, u_char *rcookie,
		struct sockaddr *sa, u_int32_t msgid)
{
    struct state *st;

    st = statetable[state_hash(icookie, rcookie, *sa)];

#ifdef DEBUG
    log(0, "find_full_state() hash %d pointer %p",
	state_hash(icookie, rcookie, *sa), st);
#endif

    while (st != (struct state *) NULL)
      if (SA_EQUAL(*sa, st->st_peer) &&
	  (!bcmp(icookie, st->st_icookie, COOKIE_SIZE)) &&
	  (!bcmp(rcookie, st->st_rcookie, COOKIE_SIZE)) &&
	  (msgid == st->st_msgid))
	break;
      else
	st = st->st_next;

    return st;
}

/*
 * Find an ISAKMP SA state object.
 */
struct state *
find_phase1_state(struct sockaddr *sa)
{
    struct state *st;
    int i;

    for (i = 0; i < STATE_TABLE_SIZE; i++)
      for (st = statetable[i]; st != (struct state *) NULL; st = st->st_next)
	if ((SA_EQUAL(*sa, st->st_peer)) &&      /* Host we want */
	    (st->st_msgid == 0) &&	        /* ISAKMP SA */
	    (st->st_protoid == PROTO_ISAKMP) &&
	    ((st->st_state == OAKLEY_MAIN_I_4) ||
	     (st->st_state == OAKLEY_MAIN_R_3)))
	  return st;

    return (struct state *) NULL;
}

/*
 * Find a state object in the hash table, without using
 * the rcookie (to find state objects where we haven't yet
 * received a message from the responder).
 */
struct state *
find_half_state(u_char *icookie, struct sockaddr *sa)
{
    struct state *st;
    u_char rcookie[COOKIE_SIZE];

    bzero(rcookie, COOKIE_SIZE);
    st = statetable[state_hash(icookie, rcookie, *sa)];

    while (st != (struct state *) NULL)
      if (SA_EQUAL(*sa, st->st_peer) &&
	  (!bcmp(icookie, st->st_icookie, COOKIE_SIZE)))
	break;
      else
	st = st->st_next;

    return st;
}
